/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;

/**
 * <p>
 * Channel Access client. A client is the top-level object provided by this
 * library and provides access to channels.
 * </p>
 * 
 * <p>
 * Several clients might run independently within one JVM, but typically you
 * should only use a single client to save resources. Separate clients are only
 * needed if you want to use different configuration options in parallel.
 * </p>
 * 
 * <p>
 * A client may keep strong or weak references to its
 * {@link ChannelAccessChannel ChannelAccessChannels}. Therefore, code should
 * always retain a strong reference to a channel as long as it wants to keep
 * that channel alive when the implementation uses weak references. However, it
 * should also destroy a channel explicitly when it does not need it any longer,
 * so that an implementation that keeps a strong reference kann still free the
 * resources associated with a channel.
 * </p>
 * 
 * <p>
 * While different implementations of a Channel Access client are possible, this
 * library only provides a single implementation, the
 * {@link DefaultChannelAccessClient}. This implementation can be configured and
 * should be suitable for most purposes.
 * </p>
 * 
 * @author Sebastian Marsching
 * 
 * @see ChannelAccessChannel
 * @see ChannelAccessClientConfiguration
 * @see DefaultChannelAccessClient
 */
public interface ChannelAccessClient {

    /**
     * Destroys this client. This method should be called if the client is not
     * needed any longer, so that its resources can be released (e.g. background
     * threads can be terminated). Destroying a client also destroys all its
     * channels and subsequently all monitors.
     */
    void destroy();

    /**
     * Returns a reference to the channel with the given name. Calling this
     * method multiple times for the same channel name will result in separate
     * instances being returned each time. In the background however, all these
     * channels will share the same implementation, so that the resource
     * overhead for the separate instances is negligible. This way, you can get
     * a channel instance (and destroy it when you do not need it any longer)
     * without having to worry whether other parts of your application still
     * need this channel.
     * 
     * @param channelName
     *            network name of the channel. The name is serialized using the
     *            configured charset (see
     *            {@link ChannelAccessClientConfiguration#getCharset()}).
     * @return new instance of the Channel Access channel with the specified
     *         name.
     * @throws IllegalStateException
     *             if this client has been destroyed.
     * @throws SerializedChannelNameTooLargeException
     *             if the serialized form of the <code>channelName</code> is too
     *             long to sent it in a search request.
     */
    ChannelAccessChannel getChannel(String channelName);

    /**
     * Returns the configuration for this Channel Access client.
     * 
     * @return configuration used by this client.
     */
    ChannelAccessClientConfiguration getConfiguration();

}
