/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon;

import java.net.InetSocketAddress;

/**
 * <p>
 * Detector for beacons sent by Channel Access servers. The beacon detector
 * listens for beacons sent by Channel Access servers and tries to detect
 * anomalies in the receive pattern. If it detects an anomaly, it notifies the
 * corresponding listeners which have been registered with the beacon detector.
 * </p>
 * 
 * <p>
 * An implementation of this interface should ensure that the methods defined by
 * this interface are thread safe and thus can be called by any thread.
 * </p>
 * 
 * <p>
 * An implementation of this interface may decide to process add and remove
 * requests for listeners asynchronously. This means, that a listener that has
 * just been added might not receive notifications immediately and that a
 * listener that has just been removed might still receive notifications for a
 * while.
 * </p>
 * 
 * <p>
 * The beacon detector may keep references to the listeners registered with it
 * in the form of strong or weak references. This means that code registering a
 * listener should always remove the listener explicitly if it is not needed any
 * longer, but should still keep a strong reference to it as long as it wants to
 * keep the listener alive.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface BeaconDetector {

    /**
     * Listener for a beacon anomaly. A beacon anomaly typically indicates that
     * a new server might have been started or a network connection might just
     * have been restored and thus a channel name resolver should try to resolve
     * channel names that it could not resolve earlier.
     * 
     * @author Sebastian Marsching
     */
    public interface BeaconAnomalyListener {

        /**
         * Called when a beacon anomaly is detected. Typically, this method will
         * be called from the communication thread of the beacon detector.
         * Therefore, this method should not perform operations that take a
         * considerable amount of time or might block.
         * 
         * @param serverAddress
         *            IP address and port number of the server that sent the
         *            beacon message which triggered the beacon anomaly (never
         *            <code>null</code>). This can be used for informational
         *            purposes.
         */
        void beaconAnomalyDetected(InetSocketAddress serverAddress);

    }

    /**
     * Listener that is notified of received beacons. This information can be
     * interesting for a component that keeps a connection to this server
     * because received beacons can be used as a health indicator. For example,
     * a regular beacon typically indicates that the connection to the server is
     * in order. An irregular beacon on the other hand might indicate that the
     * connection to the server is unreliable (e.g. it might have been
     * interrupted for a certain period or the server might have been restarted)
     * and thus it should be checked whether the connection still works or needs
     * to be reestablished.
     * 
     * @author Sebastian Marsching
     */
    public interface BeaconListener {

        /**
         * Called when a beacon is received. If the beacon detector determines
         * that the received beacon is most likely a duplicate of a beacon that
         * has been received earlier, this method is not called. Typically, this
         * method will be called from the communication thread of the beacon
         * detector. Therefore, this method should not perform operations that
         * take a considerable amount of time or might block.
         * 
         * @param serverAddress
         *            IP address and port number of the server that sent the
         *            beacon message (never <code>null</code>). This can be used
         *            to detect which server triggered the anomaly when the same
         *            listener is registered for multiple addresses.
         * @param regularBeacon
         *            <code>true</code> if this beacon was expected and arrived
         *            at the expected point in time. <code>false</code> if the
         *            beacon was unexpected (e.g. because it was the first
         *            beacon or because it was significantly delayed or early).
         *            If <code>false</code>, a component maintaining a
         *            connection to the server might want to verify that the
         *            connection is still working.
         */
        void beaconReceived(InetSocketAddress serverAddress,
                boolean regularBeacon);

    }

    /**
     * Adds beacon anomaly listener. If the listener has already been added,
     * this method does nothing. The beacon detector uses object identity, not
     * equality when deciding whether a listener has been added before. The
     * beacon detector might not keep a strong reference to the listener. This
     * means that the calling code has to keep a strong reference in order to
     * keep the listener from being garbage collected.
     * 
     * @param listener
     *            beacon anomaly listener (never <code>null</code>).
     */
    void addBeaconAnomalyListener(BeaconAnomalyListener listener);

    /**
     * Removes a beacon anomaly listener. If the listener has already been
     * removed (or has not been added before), this method does nothing. The
     * beacon detector uses object identity, not equality when deciding which
     * listener should be removed.
     * 
     * @param listener
     *            beacon anomaly listener that shall be removed (never
     *            <code>null</code>).
     */
    void removeBeaconAnomalyListener(BeaconAnomalyListener listener);

    /**
     * Adds a beacon listener. The listener is only added for the specified
     * server address. If the listener has already been added for the same
     * server address, this method does nothing. The beacon detector uses object
     * identity, not equality when deciding whether a listener has been added
     * before. The same listener can be registered for multiple server
     * addresses. The beacon detector might not keep a strong reference to the
     * listener. This means that the calling code has to keep a strong reference
     * in order to keep the listener from being garbage collected.
     * 
     * @param serverAddress
     *            IP address and port number of the server the listener is
     *            registered for (never <code>null</code>).
     * @param listener
     *            beacon listener (never <code>null</code>).
     */
    void addBeaconListener(InetSocketAddress serverAddress,
            BeaconListener listener);

    /**
     * Removes a beacon listener. The lister is only removed for the specified
     * server address. If the listener has already been removed (or has not been
     * added for this address before), this method does nothing. The beacon
     * detector uses object identity, not equality when deciding which listener
     * should be removed. If the same listener has been registered for multiple
     * server addresses, it has to be individually removed for each server
     * address.
     * 
     * @param serverAddress
     *            IP address and port number of the server the listener was
     *            originally registered for (never <code>null</code>).
     * @param listener
     *            beacon listener that shall be removed (never <code>null</code>
     *            ).
     */
    void removeBeaconListener(InetSocketAddress serverAddress,
            BeaconListener listener);

}
