/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.beacon.internal;

import java.net.InetSocketAddress;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector.BeaconListener;
import com.aquenos.epics.jackie.client.beacon.DefaultBeaconDetector;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.util.LinkedWeakIdentityHashSet;

/**
 * Beacon status for a specific Channel Access server. This class is only
 * intended for use by {@link DefaultBeaconDetector}.
 * 
 * @author Sebastian Marsching
 */
public class BeaconServerStatus {

    private long averagePeriod;
    private long beaconDetectorStartTime;
    private ErrorHandler errorHandler;
    private int lastBeaconNumber;
    private LinkedWeakIdentityHashSet<BeaconListener> listeners = new LinkedWeakIdentityHashSet<BeaconListener>();
    private int numberOfBeaconsReceived = 0;
    private InetSocketAddress serverAddress;
    private long lastTime = 0L;

    /**
     * Creates a beacon server status for the specified server.
     * 
     * @param beaconDetectorStartTime
     *            start time of the beacon detector as a Java time-stamp. This
     *            should be the time when the beacon detector had setup its
     *            socket and thus was able to receive beacons. It is internally
     *            used for calculations when trying to detect beacon anomalies.
     * @param errorHandler
     *            error handler that is used to log exceptions. This class logs
     *            exceptions that are thrown by beacon listeners to this error
     *            handler.
     * @param serverAddress
     *            IP address and port number of the server this status object is
     *            intended for.
     */
    public BeaconServerStatus(long beaconDetectorStartTime,
            ErrorHandler errorHandler, InetSocketAddress serverAddress) {
        assert (errorHandler != null);
        assert (serverAddress != null);
        this.beaconDetectorStartTime = beaconDetectorStartTime;
        this.errorHandler = errorHandler;
        this.serverAddress = serverAddress;
    }

    /**
     * Adds a beacon listener. If the listener has already been added to this
     * beacon server status, this method does nothing. The beacon server status
     * uses object identity, not equality when deciding whether a listener has
     * been added before.
     * 
     * @param listener
     *            beacon listener (never <code>null</code>).
     */
    public void addBeaconAnomalyListener(BeaconListener listener) {
        listeners.add(listener);
    }

    /**
     * Removes a beacon listener. If the listener has already been removed (or
     * has not been added before), this method does nothing. The beacon server
     * status uses object identity, not equality when deciding which listener
     * should be removed.
     * 
     * @param listener
     *            beacon listener that shall be removed (never <code>null</code>
     *            ).
     */
    public void removeBeaconAnomalyListener(BeaconListener listener) {
        listeners.remove(listener);
    }

    /**
     * Called when a beacon message for this server is received. Updates the
     * internal state and triggers the beacon listeners. Returns
     * <code>true</code> if a beacon anomaly has been detected and
     * <code>false</code> otherwise.
     * 
     * @param beaconNumberValid
     *            <code>true</code> if the <code>beaconNumber</code> can be
     *            used, <code>false</code> if it should be ignored. The beacon
     *            number might be unusable because old servers do not send one.
     * @param beaconNumber
     *            beacon sequence number for the beacon received. Only valid if
     *            <code>beaconNumberValid</code> is <code>true</code>.
     * @return <code>true</code> if a beacon anomaly is detected,
     *         <code>false</code> otherwise.
     */
    public boolean beaconReceived(boolean beaconNumberValid, int beaconNumber) {
        long currentTime = System.currentTimeMillis();

        // If this is the first beacon that we received from this server, we
        // cannot to any comparison, so we simply initialize the beacon status.
        if (numberOfBeaconsReceived == 0) {
            // Now we have received a beacon, so we can increment the number of
            // beacons received.
            numberOfBeaconsReceived++;

            // We only use the beacon number, if it is valid (the server sends
            // one).
            if (beaconNumberValid) {
                lastBeaconNumber = beaconNumber;
            }

            // The first beacon is always valid (there is no beacon number we
            // could compare), but it is not regular.
            notifyBeaconListeners(false);

            // Update the reference time so that we can compare it the next time
            // a beacon arrives.
            lastTime = currentTime;

            // We do not want to notify the general listeners. We need to
            // receive more than one beacon in order to be sure whether we have
            // a beacon anomaly.
            return false;
        }

        // We only use the beacon number, if it is valid (the server sends one).
        if (beaconNumberValid) {
            // The natural overflow and underflow behavior of Java actually has
            // the right result. If the new number is greater than the old
            // number (or wrapped to a negative number), the result is positive,
            // otherwise it is negative.
            int beaconNumberAdvance = beaconNumber - lastBeaconNumber;

            // TODO Do we always want to update the lastBeaconNumber, even if we
            // choose to ignore the beacon? The C library does it like this, but
            // this might not be the best approach.
            // Update the last received beacon number.
            lastBeaconNumber = beaconNumber;

            // If the beacon number decreased, but only by a small value, we
            // most likely saw an earlier beacon from the same server, which
            // just arrived later (multiple interfaces, different routes, etc.).
            // We ignore such a beacon.
            if (beaconNumberAdvance <= 0 && beaconNumberAdvance > -256) {
                return false;
            }

            // If the beacon number increased by more than one, but not
            // significantly, we ignore it, too. This most likely means that we
            // just saw a beacon that arrived over a different route.
            if (beaconNumberAdvance == 2 || beaconNumberAdvance == 3) {
                return false;
            }
        }

        long currentPeriod = currentTime - lastTime;

        // If this is the second beacon, we can initialize some calculated
        // values and do the first check for a beacon anomaly.
        if (numberOfBeaconsReceived == 1) {
            // Now we have received a second beacon, so we can increment the
            // number of beacons received.
            numberOfBeaconsReceived++;

            // The second beacon still is not a regular beacon, because we
            // cannot compare the periods yet.
            notifyBeaconListeners(false);

            // The average period has not been initialized yet (at least two
            // beacons are needed to calculate a period), so we initialize it
            // with the current period.
            averagePeriod = currentPeriod;

            // We can set a lower bound on the period that passed between the
            // first beacon that we saw was sent and the preceding beacon was
            // sent. If the preceding beacon had been sent while this beacon
            // detector was already running, we would have seen it.
            long firstBeaconPeriodMinimumEstimate = lastTime
                    - beaconDetectorStartTime;

            // Update the reference time so that we can compare it the next time
            // a beacon arrives.
            lastTime = currentTime;

            // If the time difference between receiving the second beacon and
            // receiving the first beacon is less than the minimum time
            // difference between the first beacon and the preceding beacon, we
            // have a beacon anomaly because the interval decreased.
            if (currentPeriod <= firstBeaconPeriodMinimumEstimate) {
                return true;
            } else {
                return false;
            }
        }

        boolean anomalyDetected;

        if (currentPeriod >= this.averagePeriod * 1.25) {
            // If the beacon is significantly late, this might be because the
            // connection was interrupted for some time and we missed a beacon.

            // This beacon is not a regular one because we missed one in
            // between or the period increased significantly (only happens
            // shortly after a server start).
            notifyBeaconListeners(false);

            // If we missed three consecutive beacons, we most likely were
            // disconnected from the server for some time. This means that we
            // have a beacon anomaly.
            if (currentPeriod >= this.averagePeriod * 3.25) {
                anomalyDetected = true;
            } else {
                anomalyDetected = false;
            }
        } else if (currentPeriod <= this.averagePeriod * 0.8) {
            // If the beacon is early, this might be because the server was
            // restarted (after a restart, the server sends beacons in a shorter
            // interval).

            // This beacon is not a regular one because it arrived too early.
            notifyBeaconListeners(false);

            // A server restart constitutes a beacon anomaly.
            anomalyDetected = true;
        } else {
            // The beacon was expected and arrived in time. This means that we
            // have a regular beacon.
            notifyBeaconListeners(true);
            anomalyDetected = false;
        }

        // We update the average period by taking 87.5 % of the old average and
        // 12.5 % of the latest period.
        averagePeriod = Math.round(currentPeriod * 0.125 + averagePeriod
                * 0.875);

        // We have to remember the current time for the next run.
        lastTime = currentTime;

        return anomalyDetected;
    }

    /**
     * Returns the last time when a beacon message has been received for this
     * server or zero if no beacon message has been received at all.
     * 
     * @return time the last beacon message was received at as a Java
     *         time-stamp.
     */
    public long getLastTime() {
        return lastTime;
    }

    /**
     * Tells whether beacon anomaly listeners are registered for this server.
     * 
     * @return <code>true</code> if there is at least one beacon anomaly
     *         listener that is registered for this server, <code>false</code>
     *         if no listeners are registered.
     */
    public boolean hasBeaconAnomalyListeners() {
        return !listeners.isEmpty();
    }

    private void notifyBeaconListeners(boolean regularBeacon) {
        for (BeaconListener listener : listeners) {
            try {
                listener.beaconReceived(serverAddress, regularBeacon);
            } catch (Throwable e) {
                errorHandler.handleError(BeaconServerStatus.class, e,
                        "A beacon listener for "
                                + serverAddress.getAddress().getHostAddress()
                                + ":" + serverAddress.getPort()
                                + " threw an exception.");
            }
        }
    }

}
