/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.LinkedList;

import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.client.ChannelAccessClientConfiguration;
import com.aquenos.epics.jackie.client.ClientThreadingStrategy;
import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.internal.ChannelAccessPVServerConnection.State;
import com.aquenos.epics.jackie.client.io.ChannelAccessServerConnection;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.util.IntegerIdPool;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Representation of a Channel Access PV server. This class is only intended for
 * use by the classes inside this package and by the
 * {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * This class is closely coupled to the {@link ChannelAccessPVServerConnection}:
 * Each server has exactly one connection and each connection has exactly one
 * server. The functionality is separated into two classes for three reasons:
 * First, both classes are already quite large, thus uniting them in a single
 * class would create an extremely large class. Second, the connection class has
 * to inherit from {@link ChannelAccessServerConnection} which would be
 * inconvenient for the {@link Server}. Third, the two objects have different
 * scopes. This server handles all requests from {@link Channel Channels} which
 * can originate in any thread. Therefore, the server uses a mutex to protect
 * access to its fields. The {@link ChannelAccessPVServerConnection} on the
 * other hand runs nearly all actions within the communication thread and
 * therefore does not need a mutex.
 * </p>
 * 
 * <p>
 * This class takes care to not hold its mutex when it notifies objects
 * (instances of {@link Channel} and {@link Subscription}). Therefore, these
 * objects can safely call methods of the server while holding their own mutex
 * without risking a dead lock. Obviously, this leaves the risk of race
 * conditions in notifications, however as virtually all notifications are
 * triggered from the communication thread, this can only happen when the server
 * is destroyed explicitly (destruction because of disconnects also runs in the
 * communication thread) and the channels and subscriptions are implemented in a
 * way that they can handle this situation.
 * </p>
 * 
 * <p>
 * A server is destroyed when its connection is destroyed or when it is
 * destroyed explicitly. When it is destroyed explicitly, it also destroys the
 * connection, which closes the associated socket channel.
 * </p>
 * 
 * <p>
 * A server keeps strong references to the {@link Channel Channels} and
 * {@link Subscription Subscriptions} registered with it, so instances of these
 * classes have to unregister themselves in order to ensure that they can be
 * garbage collected.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class Server implements CommunicationProcessor {

    private final Object lock = new Object();

    private BeaconDetector beaconDetector;
    private HashMap<Integer, Pair<Boolean, Boolean>> channelAccessRights = new HashMap<Integer, Pair<Boolean, Boolean>>();
    private HashMap<Integer, Channel> channels = new HashMap<Integer, Channel>();
    private CommunicationController communicationController;
    private ChannelAccessClientConfiguration configuration;
    private ChannelAccessPVServerConnection connection;
    private ChannelAccessVersion initialVersion;
    private HashMap<Integer, ReadOperation<ChannelAccessGettableValue<?>>> pendingReadOperations = new HashMap<Integer, ReadOperation<ChannelAccessGettableValue<?>>>();
    private HashMap<Integer, WriteOperation> pendingWriteOperations = new HashMap<Integer, WriteOperation>();
    private InetSocketAddress serverAddress;
    private ServerPool serverPool;
    private State state = State.CONNECTION_PENDING;
    private HashMap<Integer, Subscription> subscriptions = new HashMap<Integer, Subscription>();
    private ClientThreadingStrategy threadingStrategy;
    private IntegerIdPool readOperationIdPool = new IntegerIdPool(1, -1, 0L,
            false);
    private IntegerIdPool writeOperationIdPool = new IntegerIdPool(1, -1, 0L,
            false);

    /**
     * Creates a server. This only creates the object but does not establish a
     * connection. The connection is established when the
     * {@link #setCommunicationController(CommunicationController)} method is
     * called, which must be called within the communication thread.
     * 
     * @param beaconDetector
     *            beacon detector to be used by the connection associated with
     *            this server. This beacon detector is used to detect beacon
     *            anomalies which is important for the watch dog that determines
     *            whether a server is responsive or unresponsive.
     * @param configuration
     *            client configuration that shall be used by this server. The
     *            configuration is mainly used to get the {@link ErrorHandler}
     *            and to pass it on to the
     *            {@link ChannelAccessPVServerConnection}.
     * @param initialVersion
     *            initial Channel Access version to use for the connection.
     *            Typically, this version has been determined from the response
     *            send for a search request for a channel. If the connection
     *            determines the the server actually supports a newer version,
     *            it will perform a protocol version upgrade. However, it cannot
     *            perform a protocol version downgrade. In this case, the
     *            connection is destroyed.
     * @param serverAdddress
     *            socket address of the Channel Access server to which a
     *            connection shall be established.
     * @param serverPool
     *            server pool that creates this server and which will "own" the
     *            server. The server notifies this connection pool when the
     *            server is destroyed.
     * @param threadingStrategy
     *            threading strategy that is responsible for this server. The
     *            threading strategy is notified when this connection is
     *            destroyed, so that the corresponding thread can be stopped.
     *            The threading strategy is also used when notifying future
     *            completion listeners that a read or write operation has
     *            finished.
     */
    public Server(BeaconDetector beaconDetector,
            ChannelAccessClientConfiguration configuration,
            ChannelAccessVersion initialVersion,
            InetSocketAddress serverAdddress, ServerPool serverPool,
            ClientThreadingStrategy threadingStrategy) {
        this.beaconDetector = beaconDetector;
        this.configuration = configuration;
        this.initialVersion = initialVersion;
        this.serverAddress = serverAdddress;
        this.serverPool = serverPool;
        this.threadingStrategy = threadingStrategy;
    }

    private void connect() {
        try {
            ChannelAccessPVServerConnection connection = new ChannelAccessPVServerConnection(
                    beaconDetector, communicationController, configuration,
                    this, serverAddress, threadingStrategy, initialVersion);
            if (connection != null) {
                synchronized (lock) {
                    this.connection = connection;
                }
            }
        } catch (IOException e) {
            // It is save to call destroy() because we are in the communication
            // thread.
            destroy(true);
            // We do not log the exception because most likely this is just some
            // regular connection problem.
        }
    }

    /**
     * <p>
     * Registers a channel with this server. If the connection has been
     * established (responsive or unresponsive), a request is sent to the
     * Channel Access PV server to connect the channel. If the connection has
     * not been established yet, this request is sent once the connection is
     * established. If this server has been destroyed, a
     * {@link IllegalArgumentException} is thrown.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} to register itself with the
     * server.
     * </p>
     * 
     * @param channel
     *            channel that shall be registered with the server.
     */
    public void registerChannel(Channel channel) {
        synchronized (lock) {
            if (state.equals(State.CONNECTION_DESTROYED)) {
                throw new IllegalStateException(
                        "Cannot add a channel to a destroyed server.");
            }
            channels.put(channel.getCID(), channel);
            // If the connection has already been established, we connect the
            // channel. Otherwise, it will be connected when it is established.
            // We have to do this while holding the lock because this avoids a
            // state change.
            if (state.equals(State.CONNECTION_RESPONSIVE)
                    || state.equals(State.CONNECTION_UNRESPONSIVE)) {
                connection.connectChannel(channel.getCID(), channel.getName());
            }
        }
    }

    /**
     * <p>
     * Unregisters a channel from this server. This method removed the channel
     * from the map maintained by this server and sends a request to disconnect
     * the channel to the Channel Access PV server.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} to unregister itself from the
     * server when the channel is destroyed. It may only be called when this
     * server's connection has been established.
     * </p>
     * 
     * <p>
     * When the last channel belonging to this server is unregistered, the
     * server is destroyed automatically.
     * </p>
     * 
     * @param channel
     *            channel that shall be unregistered from this server.
     */
    public void unregisterChannel(Channel channel) {
        boolean noChannelsLeft;
        synchronized (lock) {
            // If the channel is not registered (it has never been registered or
            // has already been unregistered), we simply do nothing.
            if (channels.remove(channel.getCID()) == null) {
                return;
            }
            // If we have a connection and the channel is connected, we have to
            // disconnect it. If we do not have a connection, this method should
            // not have been called, because when a channel is destroyed in the
            // connecting state, it waits with calling this method until it gets
            // a connection notification.
            assert !state.equals(State.CONNECTION_PENDING);
            connection.disconnectChannel(channel.getCID(), channel.getSID());
            noChannelsLeft = channels.isEmpty();
            // If no channels are left, we change to the destroyed state. We do
            // this here (before calling the destroy method) because we want to
            // make sure that no new channels are added to this server object.
            // On the other hand, we do not want to hold the mutex while calling
            // the destroy method.
            if (noChannelsLeft) {
                state = State.CONNECTION_DESTROYED;
            }
        }
        // We call the destroy method without holding the mutex. Calling the
        // destroy methods might have a lot of side effects and we do not want
        // to risk a dead lock.
        if (noChannelsLeft) {
            // If there are no channels left, we want to destroy the connection
            // gracefully, so that the channels that have just been removed may
            // get disconnected gracefully.
            destroy(false);
        }
    }

    /**
     * <p>
     * Starts a read operation. A read operation gets a channel's current value
     * from the server. If this server is destroyed before the operation
     * finishes, the returned future is notified about the failure by calling
     * its {@link ReadOperation#operationCancelled()} method.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} to get its current value, when
     * requested.
     * </p>
     * 
     * @param channel
     *            channel to read the current value for. The channel must be
     *            connected.
     * @param dataType
     *            data type that shall be used in the response from the server.
     * @param count
     *            number of elements that shall be sent by the server. May only
     *            be zero for Channel Access version 4.13 or newer. In this case
     *            it means the channel's native count.
     * @return future that allows to check the status of the read operation.
     * @throws ChannelAccessException
     *             if the read operation cannot be started (e.g. because the
     *             server is not connected or the reply would exceed the maximum
     *             payload size).
     */
    public ReadOperation<? extends ChannelAccessGettableValue<?>> readChannel(
            Channel channel, ChannelAccessValueType dataType, int count) {
        synchronized (lock) {
            // If the server is not connected, we throw a ChannelAccessException
            // with a status code of ECA_DISCONN.
            if (!state.equals(State.CONNECTION_RESPONSIVE)
                    && !state.equals(State.CONNECTION_UNRESPONSIVE)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            // We check whether the requested data type and count will result in
            // a reply that is larger than
            IntegerIdPool.Id operationId = readOperationIdPool.acquireId();
            // It is okay to call readChannel(...) while holding the mutex
            // because the method is lock free.
            boolean success = false;
            try {
                connection.readChannel(channel.getSID(), dataType, count,
                        operationId.get());
                success = true;
            } finally {
                // If the request has not been sent, we can release the
                // operation ID immediately.
                if (!success) {
                    operationId.release();
                }
            }
            ReadOperation<ChannelAccessGettableValue<?>> operation = new ReadOperation<ChannelAccessGettableValue<?>>(
                    configuration.getErrorHandler(), operationId, dataType,
                    count, threadingStrategy);
            pendingReadOperations.put(operationId.get(), operation);
            return operation;
        }
    }

    /**
     * <p>
     * Starts a write operation. A write operation puts a value to a channel by
     * sending it to the channel's server. If this server is destroyed before
     * the operation finishes, the returned future is notified about the failure
     * by calling its {@link WriteOperation#operationCancelled()} method.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} to get update the value on the
     * server, when requested.
     * </p>
     * 
     * @param channel
     *            channel to write a new value for. The channel must be
     *            connected.
     * @param value
     *            value that shall be sent to the server.
     * @return future that allows to check the status of the write operation.
     * @throws ChannelAccessException
     *             if the read operation cannot be started (e.g. because the
     *             server is not connected or the reply would exceed the maximum
     *             payload size).
     */
    public WriteOperation writeChannel(Channel channel,
            ChannelAccessPuttableValue<?> value) {
        synchronized (lock) {
            // If the server is not connected, we throw a ChannelAccessException
            // with a status code of ECA_DISCONN.
            if (!state.equals(State.CONNECTION_RESPONSIVE)
                    && !state.equals(State.CONNECTION_UNRESPONSIVE)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            IntegerIdPool.Id operationId = writeOperationIdPool.acquireId();
            // It is okay to call writeChannel(...) while holding the mutex
            // because the method is lock free.
            boolean success = false;
            try {
                connection.writeChannel(channel.getSID(), value,
                        operationId.get());
                success = true;
            } finally {
                // If the request has not been sent, we can release the
                // operation ID immediately.
                if (!success) {
                    operationId.release();
                }
            }
            WriteOperation operation = new WriteOperation(
                    configuration.getErrorHandler(), operationId,
                    threadingStrategy);
            pendingWriteOperations.put(operationId.get(), operation);
            return operation;
        }
    }

    /**
     * <p>
     * Registers a subscription with this server. This sends a subscription
     * request to the Channel Access PV server and registers the subscription in
     * this server's internal map.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} when it creates a new
     * subscription or when the channel is connected to register the existing
     * subscriptions.
     * </p>
     * 
     * @param subscription
     *            subscription to be registered with this server.
     * @param dataType
     *            data type that shall be used in responses from the server.
     * @param count
     *            number of elements that shall be sent by the server. May only
     *            be zero for Channel Access version 4.13 or newer. In this case
     *            it means the channel's native count.
     * @throws ChannelAccessException
     *             if the subscription cannot be requested (e.g. because the
     *             server is not connected or the reply would exceed the maximum
     *             payload size).
     */
    public void registerSubscription(Subscription subscription,
            ChannelAccessValueType dataType, int count) {
        synchronized (lock) {
            // If the server is not connected, we throw a ChannelAccessException
            // with a status code of ECA_DISCONN.
            if (!state.equals(State.CONNECTION_RESPONSIVE)
                    && !state.equals(State.CONNECTION_UNRESPONSIVE)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            // Calling subscribe(...) while holding the mutex is safe because
            // that method is lock free.
            connection.subscribe(subscription.getChannel().getSID(), dataType,
                    count, subscription.getMask(),
                    subscription.getSubscriptionId().get());
            subscriptions.put(subscription.getSubscriptionId().get(),
                    subscription);
        }
    }

    /**
     * <p>
     * Unregisters a subscription from this server. This sends a subscription
     * cancellation request to the Channel Access PV server but does not remove
     * the subscription in this server's internal map. The subscription is
     * removed when the server confirms that it cancelled the subscription.
     * </p>
     * 
     * <p>
     * This method is used by a {@link Channel} when a subscription is
     * destroyed.
     * </p>
     * 
     * @param subscription
     *            subscription to be unregistered from this server.
     * @param dataType
     *            data type that has been used when requesting the subscription.
     * @param count
     *            number of elements that has been used when requesting the
     *            subscription.
     * @throws IllegalStateException
     *             if the connection to the server has not been established yet
     *             or if the server has been destroyed.
     */
    public void unregisterSubscription(Subscription subscription,
            ChannelAccessValueType dataType, int count) {
        synchronized (lock) {
            if (!state.equals(State.CONNECTION_RESPONSIVE)
                    && !state.equals(State.CONNECTION_UNRESPONSIVE)) {
                throw new IllegalStateException("Server is not connected.");
            }
            // This method should never be called for a subscription that was
            // not registered previously.
            assert subscriptions
                    .containsKey(subscription.getSubscriptionId().get());
            // Calling unsubscribe(...) while holding the mutex is safe because
            // that method is lock free.
            connection.unsubscribe(subscription.getChannel().getSID(), dataType,
                    count, subscription.getSubscriptionId().get());
            // We do not remove the subscription from the map - this will be
            // done when we receive the confirmation that the subscription has
            // been cancelled.
        }
    }

    /**
     * Returns this server's current connection state.
     * 
     * @return current connection state.
     */
    public State getState() {
        synchronized (lock) {
            return state;
        }
    }

    /**
     * <p>
     * Destroys this server. This closes the connection and notifies the
     * channels and the pending operations that the server has been destroyed.
     * </p>
     * 
     * <p>
     * If <code>immediateDisconnect</code> is <code>true</code>, the connection
     * is closed immediately. If it is <code>false</code>, the server is
     * destroyed immediately, the connection however is kept open for some time.
     * Received data is discarded, but messages that have been queued are sent.
     * The connection is closed when the send queue is empty or when a time out
     * is reached.
     * </p>
     * 
     * @param immediateDisconnect
     *            if <code>true</code> the connection is destroyed immediately,
     *            closing the communication channel with the server. If
     *            <code>false</code>, the connection is kept alive for some more
     *            time so that pending messages can be sent.
     */
    public void destroy(boolean immediateDisconnect) {
        if (immediateDisconnect) {
            // We simply destroy the actual connection. This will trigger the
            // updateConnectionState method which then takes care of notifying
            // the channels etc.
            connection.destroy();
        } else {
            // Normally, we want to send messages that have been queued before
            // closing the connection so that channels can be disconnected
            // gracefully. We update this server's connection state first so
            // that the channels are notified (and can queue their disconnect
            // messages).
            updateConnectionState(State.CONNECTION_DESTROYED);
            // After this server has been destroyed, we notify the connection
            // that it should destroy itself when the send queue has been
            // flushed.
            connection.destroyGracefully();
        }
    }

    @Override
    public void setCommunicationController(
            CommunicationController communicationController) {
        synchronized (lock) {
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            this.communicationController = communicationController;
        }
        // We know that this method is only called once. We call connect without
        // holding the mutex in order to keep the code that is protected by the
        // mutex as small as possible.
        connect();
    }

    /**
     * Notifies the server that the connection state changed. This method may
     * only be called by the {@link ChannelAccessPVServerConnection} associated
     * with this server.
     * 
     * @param newState
     *            the new connection state.
     */
    void updateConnectionState(State newState) {
        // The ChannelAccessPVServerConnection only calls this method if there
        // is a state change, therefore we can always assume that the state has
        // changed.
        LinkedList<Channel> channelsToBeConnected = new LinkedList<Channel>();
        synchronized (lock) {
            if (state.equals(State.CONNECTION_DESTROYED)) {
                // If this server has already been destroyed before, we do not
                // want to make any further state changes that might happen
                // because the connection thread triggered an event after the
                // server was destroyed but before the connection thread got
                // killed. We still want to unregister from the server pool,
                // after the connection has been destroyed, therefore we can
                // continue if the notification is about the destroyed state.
                // This is necessary, because unregisterChannel(...) sets the
                // state to destroyed but than destroys the connection which in
                // turn triggers this method. In this case, the list of channels
                // is empty and no notifications will happen.
                if (!newState.equals(State.CONNECTION_DESTROYED)) {
                    return;
                }
            }
            // When we have a transition from the connection pending to a
            // connected state, we have to connect all channels that have been
            // registered so far.
            if (state.equals(State.CONNECTION_PENDING)
                    && (newState.equals(State.CONNECTION_RESPONSIVE) || newState
                            .equals(State.CONNECTION_UNRESPONSIVE))) {
                // The connection has been established, thus we can now connect
                // the channels. We have to get the list of channels while
                // holding the mutex because channels that are added after the
                // connection has been established will be connected
                // automatically. We can queue the connection requests after
                // releasing the mutex because a disconnect request will never
                // get queued before the connect request (a channel does not
                // request disconnecting before it has been connected or the
                // connection has failed).
                channelsToBeConnected.addAll(channels.values());
            }
            state = newState;
        }
        // After releasing the mutex, we send the connect messages for the
        // channels. This list will be empty unless it has been filled
        // explicitly. Therefore, we do not have to check the state here.
        for (Channel channel : channelsToBeConnected) {
            connection.connectChannel(channel.getCID(), channel.getName());
        }
        // If the new state is the destroyed state, we have to notify the
        // channels and the server pool. We have to do this without holding the
        // mutex because these notifications might trigger more notifications
        // including user code and if we hold the mutex, there might be the risk
        // of a dead lock. This means that channels which have already
        // unregistered might be notified, but this is okay because we can
        // handle this inside the channel class.
        if (newState.equals(State.CONNECTION_DESTROYED)) {
            // We need the mutex to make a consistent copy of the list of
            // registered channels.
            LinkedList<Channel> channels = new LinkedList<Channel>();
            synchronized (lock) {
                channels.addAll(this.channels.values());
                // We clear the list of channels so that no channel gets
                // notified twice, even if the updateConnectionState(...) method
                // is called twice for some reason.
                this.channels.clear();
            }
            for (Channel channel : channels) {
                channel.serverDestroyed(this);
            }
            // Cancel pending operations. We can do this safely without holding
            // the mutex because no thread will modify the map once this server
            // has been marked as destroyed. However, this means that we cannot
            // clear the map, because this might cause a concurrent
            // modification. This should not matter, because with the server
            // destroyed, the map should not live much longer anyway.
            for (ReadOperation<? extends ChannelAccessGettableValue<?>> operation : pendingReadOperations
                    .values()) {
                operation.operationCancelled();
            }
            for (WriteOperation operation : pendingWriteOperations.values()) {
                operation.operationCancelled();
            }
            // We have to notify the server pool. This ensures that this server
            // is removed from the pool.
            serverPool.serverDestroyed(serverAddress, this);
            // We do not have to notify the subscriptions. They will be notified
            // by their channel (if a notification is needed).
        }
    }

    /**
     * Notifies the server that the connection watch-dog has determined that the
     * connection is now responsive. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     */
    void connectionResponsive() {
        // We need the mutex to make a consistent copy of the list of registered
        // channels and subscriptions.
        LinkedList<Channel> channels = new LinkedList<Channel>();
        LinkedList<Subscription> subscriptions = new LinkedList<Subscription>();
        synchronized (lock) {
            channels.addAll(this.channels.values());
            subscriptions.addAll(this.subscriptions.values());
        }
        for (Channel channel : channels) {
            channel.serverResponsive(this);
        }
        // The a connection is responsive (again), we want to notify all
        // subscriptions, because they might want to request a value update
        // because they might have missed events.
        for (Subscription subscription : subscriptions) {
            subscription.serverResponsive(this);
        }
    }

    /**
     * Notifies the server that the connection watch-dog has determined that the
     * connection is now unresponsive. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     */
    void connectionUnresponsive() {
        // We need the mutex to make a consistent copy of the list of registered
        // channels.
        LinkedList<Channel> channels = new LinkedList<Channel>();
        synchronized (lock) {
            channels.addAll(this.channels.values());
        }
        for (Channel channel : channels) {
            channel.serverUnresponsive(this);
        }
    }

    /**
     * Notifies the server that a message indicating that a channel has been
     * connected has been received. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param cid
     *            client-assigned ID identifying the channel. This is the ID
     *            that was specified when connecting the channel.
     * @param sid
     *            server-assigned ID identifying the channel. This is the ID
     *            that needs to be used when sending requests regarding the
     *            channel to the server.
     * @param nativeDataType
     *            native data-type of the channel. This is the data-type that
     *            the server can support without having to make any conversions.
     * @param nativeCount
     *            native element count of the channel. This is the number of
     *            elements that values for the channel have on the server side.
     */
    void channelConnected(int cid, int sid,
            ChannelAccessValueType nativeDataType, int nativeCount) {
        Channel channel;
        boolean accessRightsValid = false;
        boolean mayRead = false;
        boolean mayWrite = false;
        synchronized (lock) {
            // We get the access rights received earlier and remove them at the
            // same time because we do not need them any longer.
            Pair<Boolean, Boolean> accessRights = channelAccessRights
                    .remove(cid);
            if (accessRights != null) {
                accessRightsValid = true;
                mayRead = accessRights.getLeft();
                mayWrite = accessRights.getRight();
            }
            channel = channels.get(cid);
            // A channel is never unregistered before it has been connected (or
            // the connection failed). Therefore, we can assume that the channel
            // is never null.
            assert (channel != null);
        }
        // We do not want to hold the mutex when notifying the channel because
        // this can have a lot of side effects including the execution of
        // user-defined code and thus there is the risk of a dead lock.
        channel.connected(this, sid, nativeDataType, nativeCount,
                accessRightsValid, mayRead, mayWrite);
    }

    /**
     * Notifies the server that information about access rights have been
     * received for a channel. Access rights should be received right before a
     * notification about a connected channel is received. This means that this
     * method is typically called directly before
     * {@link #channelConnected(int, int, ChannelAccessValueType, int)} is
     * called. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param cid
     *            client-assigned ID identifying the channel. This is the ID
     *            that was specified when connecting the channel.
     * @param mayRead
     *            <code>true</code> if the channel allows read (get, monitor)
     *            operations. <code>false</code> if it does not allow such
     *            operations.
     * @param mayWrite
     *            <code>true</code> if the channel allows write (put)
     *            operations. <code>false</code> if it does not allow such
     *            operations.
     */
    void receivedAccessRights(int cid, boolean mayRead, boolean mayWrite) {
        synchronized (lock) {
            channelAccessRights.put(cid, Pair.of(mayRead, mayWrite));
        }
    }

    /**
     * Notifies the server that a connection attempt for a channel failed. This
     * kind of notification indicates that the server send a message that it
     * could not connect the requested channel in response to a connection
     * request.
     * 
     * @param cid
     *            client-assigned ID identifying the channel. This is the ID
     *            that was specified in the connection request.
     */
    void connectChannelFailed(int cid) {
        Channel channel;
        synchronized (lock) {
            channel = channels.remove(cid);
            // Typically, we should not have received any access rights if the
            // connection failed, but it is better to be sure that we do not
            // have a potential memory leak.
            channelAccessRights.remove(cid);
        }
        // We do not want to hold the mutex when notifying the channel because
        // this can have a lot of side effects including the execution of
        // user-defined code and thus there is the risk of a dead lock. This
        // means that a channel which has already been unregistered might be
        // notified, but this is okay because we can handle this inside the
        // channel class.
        if (channel != null) {
            channel.connectFailed(this);
        }
    }

    /**
     * Notifies the server that a channel has been disconnected by a server.
     * Such a notification might occur at any time after a channel has
     * successfully been connected. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param cid
     *            client-assigned ID identifying the channel. This is the ID
     *            that was specified in the connection request.
     */
    void channelDisconnectedByServer(int cid) {
        Channel channel;
        synchronized (lock) {
            channel = channels.remove(cid);
            // Typically, we should only receive this message if the channel has
            // already been connected. Therefore, the following line should be
            // unnecessary, but it is better to be safe than sorry.
            channelAccessRights.remove(cid);
        }
        // We do not want to hold the mutex when notifying the channel because
        // this can have a lot of side effects including the execution of
        // user-defined code and thus there is the risk of a dead lock. This
        // means that a channel which has already been unregistered might be
        // notified, but this is okay because we can handle this inside the
        // channel class.
        if (channel != null) {
            channel.disconnectedByServer(this);
        }
    }

    /**
     * Notifies the server that a response to a read request has been received.
     * This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param operationId
     *            client-assigned operation ID that was used when requesting the
     *            read operation.
     * @param status
     *            status of the read operation. If the status message is not
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL}, the
     *            value is invalid (or <code>null</code>) and an error occurred
     *            on the server while trying to read the channel's value.
     * @param value
     *            value read from the channel. Only valid if the
     *            <code>status</code> indicates success.
     */
    void readResponseReceived(int operationId, ChannelAccessStatus status,
            ChannelAccessGettableValue<?> value) {
        ReadOperation<? extends ChannelAccessGettableValue<?>> operation;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            // We have to hold the lock while modifying the map, however we do
            // not want to hold the mutex when triggering event listeners
            // because this might cause a dead lock.
            operation = pendingReadOperations.remove(operationId);
        }
        if (operation == null) {
            // The operation was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        try {
            if (value != null && status.matchesMessage(
                    ChannelAccessStatus.StatusMessage.ECA_NORMAL)) {
                // The read operation completed successfully. However, we need
                // to verify that the data type and count match the request.
                if (!value.getType().equals(operation.getRequestedDataType())
                        || (value.getValueSize() != operation
                                .getRequestedCount()
                                && operation.getRequestedCount() != 0)) {
                    operation.setExecutionException(new ChannelAccessException(
                            ChannelAccessStatus.StatusMessage.ECA_GETFAIL
                                    .toStatus()));
                } else {
                    // We have to do an unchecked cast in order to set the
                    // result. We can do this because we verified that the type
                    // matches and thus we can sure that we never end up with an
                    // invalid cast.
                    @SuppressWarnings("unchecked")
                    ReadOperation<ChannelAccessGettableValue<?>> untypedOperation = (ReadOperation<ChannelAccessGettableValue<?>>) operation;
                    untypedOperation.setResult(value);
                }
            } else {
                operation.setExecutionException(
                        new ChannelAccessException(status));
            }
        } catch (IllegalStateException e) {
            // The call to setResult(...) or setExecutionException(...) might
            // fail because the operation might already have been cancelled.
            // This is not considered an error.
        }
        // Now that we have received a response for the operation, it is safe to
        // reuse the operation ID.
        operation.getOperationId().release();
    }

    /**
     * Notifies the server that an error message was received in response to a
     * read request. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param operationId
     *            client-assigned operation ID that was used when requesting the
     *            read operation.
     * @param status
     *            Channel Access status code associated with the error.
     * @param extendedErrorMessage
     *            textual representation of the error sent by the server or
     *            <code>null</code> if no such message was included in the error
     *            message.
     */
    void readErrorReceived(int operationId, ChannelAccessStatus status,
            String extendedErrorMessage) {
        ReadOperation<? extends ChannelAccessGettableValue<?>> operation;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            // We have to hold the lock while modifying the map, however we do
            // not want to hold the mutex when triggering event listeners
            // because this might cause a dead lock.
            operation = pendingReadOperations.remove(operationId);
        }
        if (operation == null) {
            // The operation was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        try {
            operation.setExecutionException(
                    new ChannelAccessException(status, extendedErrorMessage));
        } catch (IllegalStateException e) {
            // The call to setResult(...) or setExecutionException(...) might
            // fail because the operation might already have been cancelled.
            // This is not considered an error.
        }
        // Now that we have received a response for the operation, it is safe to
        // reuse the operation ID.
        operation.getOperationId().release();
    }

    /**
     * Notifies the server that a response to a write request has been received.
     * This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param operationId
     *            client-assigned operation ID that was used when requesting the
     *            write operation.
     * @param status
     *            status of the read operation. If the status message is not
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL}, the
     *            write operation failed.
     */
    void writeResponseReceived(int operationId, ChannelAccessStatus status) {
        WriteOperation operation;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            // We have to hold the lock while modifying the map, however we do
            // not want to hold the mutex when triggering event listeners
            // because this might cause a dead lock.
            operation = pendingWriteOperations.remove(operationId);
        }
        if (operation == null) {
            // The operation was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // could do.
            return;
        }
        try {
            if (status.matchesMessage(
                    ChannelAccessStatus.StatusMessage.ECA_NORMAL)) {
                // The write operation completed successfully. We simply notify
                // the future. This is done by setting a null value.
                operation.setResult(null);
            } else {
                operation.setExecutionException(
                        new ChannelAccessException(status));
            }
        } catch (IllegalStateException e) {
            // The call to setResult(...) or setExecutionException(...) might
            // fail because the operation might already have been cancelled.
            // This is not considered an error.
        }
        // Now that we have received a response for the operation, it is safe to
        // reuse the operation ID.
        operation.getOperationId().release();
    }

    /**
     * Notifies the server that an error message was received in response to a
     * write request. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param operationId
     *            client-assigned operation ID that was used when requesting the
     *            write operation.
     * @param status
     *            Channel Access status code associated with the error.
     * @param extendedErrorMessage
     *            textual representation of the error sent by the server or
     *            <code>null</code> if no such message was included in the error
     *            message.
     */
    void writeErrorReceived(int operationId, ChannelAccessStatus status,
            String extendedErrorMessage) {
        WriteOperation operation;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            // We have to hold the lock while modifying the map, however we do
            // not want to hold the mutex when triggering event listeners
            // because this might cause a dead lock.
            operation = pendingWriteOperations.remove(operationId);
        }
        if (operation == null) {
            // The operation was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        try {
            operation.setExecutionException(
                    new ChannelAccessException(status, extendedErrorMessage));
        } catch (IllegalStateException e) {
            // The call to setResult(...) or setExecutionException(...) might
            // fail because the operation might already have been cancelled.
            // This is not considered an error.
        }
        // Now that we have received a response for the operation, it is safe to
        // reuse the operation ID.
        operation.getOperationId().release();
    }

    /**
     * Notifies the server that a subscription event has been received. A
     * subscription event is sent by the server after successfully creating a
     * subscription and every time that an event for that subscription occurs.
     * This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param subscriptionId
     *            client-assigned subscription ID that was used when requesting
     *            the subscription.
     * @param status
     *            Channel Access status associated with the event. If the status
     *            message is not
     *            {@link ChannelAccessStatus.StatusMessage#ECA_NORMAL}, the
     *            value is invalid (or <code>null</code>) and an error occurred
     *            on the server while trying to read the channel's value.
     * @param value
     *            value read from the channel and associated with the event.
     *            Only valid if the <code>status</code> indicates success.
     */
    void subscriptionEventReceived(int subscriptionId,
            ChannelAccessStatus status, ChannelAccessGettableValue<?> value) {
        Subscription subscription;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            subscription = subscriptions.get(subscriptionId);
        }
        if (subscription == null) {
            // The subscription was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        subscription.eventReceived(this, status, value);
    }

    /**
     * Notifies the server that a subscription has been cancelled. Typically,
     * such a notification is sent in reply to a request to cancel a
     * subscription. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param subscriptionId
     *            client-assigned subscription ID that was used when requesting
     *            (and canceling) the subscription.
     */
    void subscriptionCancelled(int subscriptionId) {
        Subscription subscription;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            subscription = subscriptions.remove(subscriptionId);
        }
        if (subscription == null) {
            // The subscription was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        subscription.cancelled(this);
    }

    /**
     * Notifies the server that an attempt to create a subscription failed and
     * an error message was sent in reply to this request. The subscription has
     * not been created. This method may only be called by the
     * {@link ChannelAccessPVServerConnection} associated with this server and
     * is always called within the communication thread.
     * 
     * @param subscriptionId
     *            client-assigned subscription ID that was used when requesting
     *            the subscription.
     * @param status
     *            Channel Access status code associated with the error.
     * @param extendedErrorMessage
     *            textual representation of the error sent by the server or
     *            <code>null</code> if no such message was included in the error
     *            message.
     */
    void subscriptionErrorReceived(int subscriptionId,
            ChannelAccessStatus status, String extendedErrorMessage) {
        Subscription subscription;
        synchronized (lock) {
            // If this server has already been destroyed, we do not want to
            // handle the response and simply ignore it.
            if (state.equals(State.CONNECTION_DESTROYED)) {
                return;
            }
            // If we receive an error in response to the subscription request,
            // the subscription has not been registered on the server and we
            // should remove it from the map.
            subscription = subscriptions.remove(subscriptionId);
        }
        if (subscription == null) {
            // The subscription was already removed. Usually, this should not
            // happen unless the server sends duplicate responses. Anyway, we
            // simply ignore the response because there is nothing else that we
            // we could do.
            return;
        }
        subscription.errorReceived(this, status, extendedErrorMessage);
    }

}
