/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.exception.InvalidPriorityException;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.exception.UnsupportedProtocolVersionException;
import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;

/**
 * Representation of a <code>CA_PROTO_CREATE_CHAN</code> message that is sent
 * from a client to a server. This message is sent in order to connect to a
 * channel. Multiple channels can be connected within the same TCP connection.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_CREATE_CHAN
 * @see ChannelAccessConnectChannelServerMessage
 * @see ChannelAccessConnectChannelFailedMessage
 */
public final class ChannelAccessConnectChannelClientMessage extends
        ChannelAccessConnectChannelMessage {

    private int cid;
    private ChannelAccessVersion version;
    private String channelName;

    /**
     * Creates a <code>CA_PROTO_CREATE_CHAN</code> message. This constructor
     * should be used to create a message that is supposed to be sent from a
     * client to a server.
     * 
     * @param cid
     *            channel CID used by the client. This is the identifier that
     *            the client will use to identify the channel in the response
     *            from the server. This number might be negative or zero.
     *            However, the value 0xffffffff is reserved for special use and
     *            should be avoided.
     * @param version
     *            Channel Access version supported by the client.
     * @param channelName
     *            the name of the channel that shall be connected.
     * @throws IllegalArgumentException
     *             if the <code>channelCID</code> is 0xffffffff.
     */
    public ChannelAccessConnectChannelClientMessage(int cid,
            ChannelAccessVersion version, String channelName) {
        super();

        if (cid == 0xffffffff) {
            throw new IllegalArgumentException(
                    "The channel CID 0xffffffff is reserved for special purposes and must not be used when connecting a channel.");
        }
        if (version == null) {
            throw new NullPointerException(
                    "Channel Acccess version must not be null.");
        }

        if (channelName == null) {
            throw new NullPointerException("Channel name must not be null.");
        }

        this.cid = cid;
        this.version = version;
        this.channelName = channelName;
    }

    private ChannelAccessConnectChannelClientMessage(
            ChannelAccessMessageHeader messageHeader, int cid,
            ChannelAccessVersion version, String channelName) {
        super(messageHeader);

        this.cid = cid;
        this.version = version;
        this.channelName = channelName;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client will use to identify the channel in the response from the
     * server. This number might be negative or zero. The value 0xffffffff is
     * reserved for special use and should be avoided. However, this value might
     * still be used in messages sent from a client.
     * 
     * @return channel CID assigned to the channel by the client.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the Channel Access version supported by this client. A server can
     * use this information in order to decide whether to reply to a search
     * request. For example, a server might not reply to a search request if the
     * protocol version supported by the client is too old for the server.
     * 
     * @return Channel Access version supported by the client (never
     *         <code>null</code>).
     */
    public ChannelAccessVersion getVersion() {
        return version;
    }

    /**
     * Returns the name of the channel that the client wants to connect to.
     * 
     * @return name of the channel that shall be connected. Might be
     *         <code>null</code> if this message instance has been reconstructed
     *         without its payload (as part of the payload of an error message).
     */
    public String getChannelName() {
        return channelName;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "Connections to servers that do not at least support Channel Access version "
                            + ChannelAccessVersion.V4_4
                            + " are not supported by this library.");
        }
        if (channelName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        Pair<byte[], Integer> channelNameBytesWithLength = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(channelName,
                        ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH,
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT, charset);
        if (channelNameBytesWithLength == null) {
            throw new SerializedChannelNameTooLargeException(
                    "The serialized form of the channel name must not be longer than "
                            + ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH
                            + " bytes.");
        }
        byte[] channelNameAsBytes = channelNameBytesWithLength.getLeft();
        int channelNameLength = channelNameBytesWithLength.getRight();
        serializeHeader(byteSink, channelNameLength, (short) 0, 0, cid,
                this.version.getMinorVersionAsShort(), version, maxPayloadSize);
        byteSink.putByteArray(channelNameAsBytes, 0, channelNameLength);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        if (!version.includes(ChannelAccessVersion.V4_4)) {
            throw new UnsupportedProtocolVersionException(
                    "Connections to servers that do not at least support Channel Access version "
                            + ChannelAccessVersion.V4_4
                            + " are not supported by this library.");
        }
        if (channelName == null) {
            throw new UnsupportedOperationException(
                    "Cannot serialize the message payload for a message that has been constructed without the payload.");
        }
        Pair<byte[], Integer> channelNameBytesWithLength = NullTerminatedStringUtil
                .stringToSizeLimitedNullTerminatedBytesOrNull(channelName,
                        ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH,
                        ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT, charset);
        if (channelNameBytesWithLength == null) {
            throw new SerializedChannelNameTooLargeException(
                    "The serialized form of the channel name must not be longer than "
                            + ChannelAccessConstants.MAX_CHANNEL_NAME_BYTES_LENGTH
                            + " bytes.");
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessConnectChannelClientMessage other = (ChannelAccessConnectChannelClientMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.version, other.version)
                .append(this.channelName, other.channelName).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(version).append(channelName).toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", version=");
        sb.append(version.toString());
        sb.append(", channelName=");
        if (channelName != null) {
            sb.append('"');
            sb.append(StringEscapeUtils.escapeJava(channelName));
            sb.append('"');
        } else {
            sb.append("null");
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @param charset
     *            encoding to use for deserializing the channel name. Usually,
     *            channel names should consist of ASCII characters only, however
     *            this is not strictly enforced by the protocol, so the encoding
     *            might matter.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     * @throws InvalidPriorityException
     *             if the priority specified in the message header is out of
     *             range.
     */
    protected static ChannelAccessConnectChannelClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly, Charset charset) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_CREATE_CHAN
                .getCommandNumber());
        String channelName;
        if (!headerOnly) {
            byte[] channelNameAsBytes = byteSource.getByteArray(messageHeader
                    .getPayloadSize());
            channelName = NullTerminatedStringUtil.nullTerminatedBytesToString(
                    channelNameAsBytes, charset);
        } else {
            channelName = null;
        }
        return new ChannelAccessConnectChannelClientMessage(messageHeader,
                messageHeader.getCID(),
                ChannelAccessVersion.forMinorVersion(messageHeader
                        .getContextSpecific()), channelName);
    }

}
