/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

/**
 * Raw header of a Channel Access message. This header already includes handling
 * for the extended (large) header format, but the data stored in instances of
 * this class has to be considered unverified, as it is an exact copy of the
 * data received from the network.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessMessageHeader {

    private short command;
    private int payloadSize;
    private short dataType;
    private int count;
    private int cid;
    private int contextSpecific;
    private boolean largeHeader;

    /**
     * Creates a CA message header, using the supplied parameters to initialize
     * the header fields. This class is only intended to be internally used by
     * Channel Access messages and thus this constructor is package private.
     * 
     * @param command
     *            "command" field.
     * @param payloadSize
     *            "postsize" field. For an extended (large) header, this should
     *            be the "postsize" from the extended part of the header.
     * @param dataType
     *            "dataType" field.
     * @param count
     *            "count" field. For an extended (large) header, this should be
     *            the "count" from the extended part of the header.
     * @param cid
     *            "cid" field.
     * @param contextSpecific
     *            "available" field.
     * @param largeHeader
     *            <code>true</code> if the message header is a "large" header
     *            (with large payload size and count fields), <code>false</code>
     *            if it is a standard header.
     */
    ChannelAccessMessageHeader(short command, int payloadSize, short dataType,
            int count, int cid, int contextSpecific, boolean largeHeader) {
        this.command = command;
        this.payloadSize = payloadSize;
        this.dataType = dataType;
        this.count = count;
        this.cid = cid;
        this.contextSpecific = contextSpecific;
        this.largeHeader = largeHeader;
    }

    /**
     * Returns the value of the header's "command" field.
     * 
     * @return "command" field.
     */
    public short getCommand() {
        return command;
    }

    /**
     * Returns the value of the header's "postsize" field. For a header in the
     * extended format, this is the field from the extended part of the header.
     * 
     * @return "postsize" field.
     */
    public int getPayloadSize() {
        return payloadSize;
    }

    /**
     * Returns the value of the header's "dataType" field.
     * 
     * @return "dataType" field.
     */
    public short getDataType() {
        return dataType;
    }

    /**
     * Returns the value of the header's "count" field. For a header in the
     * extended format, this is the field from the extended part of the header.
     * 
     * @return "count" field.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the value of the header's "cid" field.
     * 
     * @return "cid" field.
     */
    public int getCID() {
        return cid;
    }

    /**
     * Returns the value of the header's "available" field.
     * 
     * @return "available" field.
     */
    public int getContextSpecific() {
        return contextSpecific;
    }

    /**
     * <p>
     * Returns a hash code value for this message header. This method is
     * supported for the benefit of hash tables. The hash code is calculated as
     * a combination of the hash codes of the header fields, including the
     * command and the header size (even though the latter is technically not a
     * header field).
     * </p>
     * 
     * <p>
     * If two header are equal (as indicated by the return value of
     * {@link #equals(Object)}), they also have the same hash code. However, the
     * fact that two header have the same hash code does not imply that they are
     * equal.
     * </p>
     * 
     * @return hash code suitable for use by a hash table.
     */
    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(command).append(payloadSize)
                .append(dataType).append(count).append(cid)
                .append(contextSpecific).append(largeHeader).toHashCode();
    }

    /**
     * <p>
     * Tells whether the specified object represents the same message header.
     * </p>
     * 
     * <p>
     * Two headers are considered equal if they are of the same type and have
     * the same headers and payload. For comparing the types, the actual Java
     * type and not just the Channel Access command is compared.
     * </p>
     * 
     * @param obj
     *            reference object with which this object is compared.
     * @return <code>true</code> if <code>obj</code> is a message that is of the
     *         same type and has the same headers and payload as this message.
     *         <code>false</code> if <code>obj</code> is <code>null</code> or of
     *         a different type than this message or differs in the headers or
     *         payload.
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof ChannelAccessMessageHeader)) {
            return false;
        }
        ChannelAccessMessageHeader other = (ChannelAccessMessageHeader) obj;
        return new EqualsBuilder().append(this.command, other.command)
                .append(this.payloadSize, other.payloadSize)
                .append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.cid, other.cid)
                .append(this.contextSpecific, other.contextSpecific)
                .append(this.largeHeader, other.largeHeader).isEquals();
    }

    /**
     * Returns a string representation of this message header. The string
     * representation includes all header fields.
     * 
     * @return string representation of this message header.
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("CA header (");
        sb.append("cmmd=");
        sb.append(command);
        sb.append(", postsize=");
        sb.append(payloadSize);
        sb.append(", dataType=");
        sb.append(dataType);
        sb.append(", count=");
        sb.append(count);
        sb.append(", cid=");
        sb.append(cid);
        sb.append(", available=");
        sb.append(contextSpecific);
        sb.append(")");
        return sb.toString();
    }

    /**
     * Returns the size of the serialized form of this header in bytes.
     * 
     * @return size of this message header in bytes.
     */
    public int getHeaderSize() {
        if (largeHeader) {
            return ChannelAccessConstants.LARGE_HEADER_SIZE;
        } else {
            return ChannelAccessConstants.STANDARD_HEADER_SIZE;
        }
    }

}
