/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

import java.nio.charset.Charset;
import java.util.Collection;
import java.util.List;

/**
 * <p>
 * Channel Access value that represents the
 * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} type. This
 * type can only be used in get (read) operations and provides a way to check a
 * channel's alarm acknowledgement status. This information includes the
 * channel's current alarm severity, the severity of the highest unacknowledged
 * alarm and the current status of the acknowledge transient alarms flag. In
 * addition to that, a string representation of the channels current value is
 * included.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @see ChannelAccessAcknowledgeAlarm
 * @see ChannelAccessConfigureAcknowledgeTransientAlarms
 * @author Sebatian Marsching
 */
public interface ChannelAccessAlarmAcknowledgementStatus extends
        ChannelAccessGettableValue<String> {

    /**
     * <p>
     * Returns the list storing the actual value elements. The list is returned
     * by reference, so subsequent changes to the list will affect this value
     * and vice-versa. Therefore, the list may not be modified after this value
     * has been passed to a different component. If this value is read-only, the
     * list returned is also read-only.
     * </p>
     * 
     * <p>
     * The list returned does not allow the insertion of <code>null</code>
     * elements.
     * </p>
     * 
     * @return list storing this value's elements (never <code>null</code>).
     *         Elements of the list are never <code>null</code>.
     * @see #getValueSize()
     */
    List<String> getValue();

    /**
     * <p>
     * Replaces this value's elements with the elements from the specified
     * collection. This has the same effect as calling <code>clear()</code>
     * followed by <code>addAll(value)</code> on the list returned by
     * {@link #getValue()}.
     * </p>
     * 
     * <p>
     * The string's raw representation is updated automatically. Strings that
     * are truncated while being converted to their raw representation still
     * maintain their original length in the list of element strings.
     * </p>
     * 
     * @param value
     *            collection of strings that should replace this value's
     *            elements.
     * @throws NullPointerException
     *             if <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setValue(Collection<String> value);

    /**
     * <p>
     * Returns the raw representation of the strings. This an array of N * 40
     * bytes, where N is the number of strings (as returned by
     * {@link #getValueSize()}). Each string in this array has exactly 40 bytes
     * and is null-terminated. This means that the string with the zero-based
     * index i starts at an array index of i * 40.
     * </p>
     * 
     * <p>
     * The array returned is a copy, so modifications to it will not affect this
     * value.
     * </p>
     * 
     * @return raw representation of the string values (never <code>null</code>
     *         ).
     * @see #getValue()
     * @see #getValueSize()
     * @see #getCharset()
     */
    byte[] getRawValue();

    /**
     * <p>
     * Updates the string values using the specified raw representations. The
     * size of the array passed must be an integer multiple of 40. Each sequence
     * of 40 bytes is interpreted as one null-terminated string. The strings
     * must be null-terminated, therefore the size of an individual string must
     * not exceed 39 characters (not including the terminating null-byte).
     * </p>
     * 
     * <p>
     * The array is copied, thus subsequent modifications of the array will not
     * affect this value. The {@link String Java strings} returned by
     * {@link #getValue()} are updated by converting the specified raw strings
     * using the specified {@link #getCharset() encoding}.
     * </p>
     * 
     * @param rawValue
     *            new raw representation of the strings (never <code>null</code>
     *            ). The size of the array must be a multiple of 40 and strings
     *            must be null terminated.
     * @throws IllegalArgumentException
     *             if the size of the array is not an integer multiple of 40 or
     *             the individual strings are not null terminated.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setRawValue(byte[] rawValue);

    /**
     * Returns the alarm status of the value. The alarm status provides
     * information about the cause of an alarm (e.g. value too low or too high,
     * invalid value, etc.).
     * 
     * @return alarm status of the value (never <code>null</code>).
     */
    ChannelAccessAlarmStatus getAlarmStatus();

    /**
     * Set the alarm status of the value. The alarm status provides information
     * about the cause of an alarm (e.g. value too low or too high, invalid
     * value, etc.).
     * 
     * @param alarmStatus
     *            new alarm status of the value (never <code>null</code>).
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setAlarmStatus(ChannelAccessAlarmStatus alarmStatus);

    /**
     * Returns the current alarm severity of the value. The alarm severity
     * signals how severe an alarm situation is.
     * 
     * @return current alarm severity of the value (never <code>null</code>).
     */
    ChannelAccessAlarmSeverity getAlarmSeverity();

    /**
     * Sets the alarm severity of the value. The alarm severity signals how
     * severe an alarm situation is.
     * 
     * @param alarmSeverity
     *            new current alarm severity of the value (never
     *            <code>null</code>).
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setAlarmSeverity(ChannelAccessAlarmSeverity alarmSeverity);

    /**
     * Returns a flag indicating whether transient alarms have to be
     * acknowledged explicitly (<code>true</code>) or are automatically
     * acknowledged when they disappear(<code>false</code>).
     * 
     * @return whether transient alarms have to be acknowledged explicitly.
     */
    boolean isExplicitlyAcknowledgeTransientAlarms();

    /**
     * Sets the flag indicating whether transient alarms have to be acknowledged
     * explicitly (<code>true</code>) or are automatically acknowledged when
     * they disappear.
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            <code>true</code> if transient alarms have to be acknowledged
     *            explicitly, <code>false</code> if they are automatically
     *            acknowledged when they disappear.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setExplicitlyAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms);

    /**
     * Returns the severity of the highest unacknowledged alarm. If the numeric
     * value that represents the severity is greater than
     * {@link ChannelAccessAlarmSeverity#INVALID_ALARM INVALID_ALARM}, the
     * <code>INVALID_ALARM</code> severity is returned. Technically, a negative
     * number is interpreted as a higher severity, because in the protocol
     * specification, the severity is represented by an unsigned integer.
     * 
     * @return severity of the highest unacknowledged alarm (never
     *         <code>null</code>).
     */
    ChannelAccessAlarmSeverity getUnacknowledgedAlarmSeverity();

    /**
     * Sets the severity of the highest unacknowledged alarm.
     * 
     * @param unacknowledgedAlarmSeverity
     *            severity of the highest unacknowledged alarm (never
     *            <code>null</code>).
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setUnacknowledgedAlarmSeverity(
            ChannelAccessAlarmSeverity unacknowledgedAlarmSeverity);

    /**
     * Returns the encoding that is used for converting {@linkplain String Java
     * strings} into raw (null-terminated) strings and vice-versa. The encoding
     * is defined when a value instance is created and never changes.
     * 
     * @return encoding used for encoding and decoding strings (never
     *         <code>null</code>).
     */
    Charset getCharset();

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessAlarmAcknowledgementStatus asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessAlarmAcknowledgementStatus clone();

}
