/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

/**
 * Alarm severity associated with a Channel Access channel.
 * 
 * @author Sebastian Marsching
 */
public enum ChannelAccessAlarmSeverity {

    /**
     * No alarm. Indicates that everything is OK.
     */
    NO_ALARM((short) 0),

    /**
     * Minor alarm. Typically, this means that a value is not within the
     * expected range, but no immediate action is required.
     */
    MINOR_ALARM((short) 1),

    /**
     * Major Alarm. This severity typically indicates that a parameter is in a
     * range that indicates a malfunction and actions are needed.
     */
    MAJOR_ALARM((short) 2),

    /**
     * Unknown alarm severity. Typically, this is caused by a parameter being
     * out of date (e.g. because it could not be read from the hardware). This
     * is the highest severity, because the actual severity cannot be determined
     * and thus the worst has to be expected.
     */
    INVALID_ALARM((short) 3);

    private short severityCode;

    private ChannelAccessAlarmSeverity(short severityCode) {
        this.severityCode = severityCode;
    }

    /**
     * Returns the severity code associated with this severity. This is an
     * integer that uniquely identifies this severity. This number is suitable
     * to be sent over the network.
     * 
     * @return integer number between 0 and 3, identifying this severity.
     * @see #forSeverityCode(short)
     */
    public short toSeverityCode() {
        return severityCode;
    }

    /**
     * Returns the severity for the given <code>severityCode</code>. This can be
     * used to translate a severity code received from the network into a
     * severity.
     * 
     * @param severityCode
     *            integer number identifying the severity. Must be between 0 and
     *            3.
     * @return the severity identified by the <code>severityCode</code>, never
     *         <code>null</code>.
     * @throws IllegalArgumentException
     *             if the <code>severityCode</code> does not represent a valid
     *             severity (is not 0, 1, 2, or 3).
     */
    public static ChannelAccessAlarmSeverity forSeverityCode(short severityCode) {
        switch (severityCode) {
        case 0:
            return NO_ALARM;
        case 1:
            return MINOR_ALARM;
        case 2:
            return MAJOR_ALARM;
        case 3:
            return INVALID_ALARM;
        default:
            throw new IllegalArgumentException(
                    "Severity code must be 0, 1, 2, or 3.");
        }
    };

}
