/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value.internal;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;
import java.util.Collection;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessClassName;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Implementation of DBR_CLASS_NAME.
 * 
 * @author Sebastian Marsching
 */
public final class ChannelAccessClassNameImpl extends
        AbstractStringBase<ChannelAccessClassNameImpl> implements
        ChannelAccessClassName {

    /**
     * Creates a DBR_CLASS_NAME.
     * 
     * @param value
     *            string value.
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    public ChannelAccessClassNameImpl(Collection<String> value, Charset charset) {
        super(value, charset);
    }

    /**
     * Creates a read-only DBR_CLASS_NAME. All read operations are delegated to
     * <code>delegate</code>.
     * 
     * @param delegate
     *            value to which read operations are delegated.
     */
    private ChannelAccessClassNameImpl(ChannelAccessClassNameImpl delegate) {
        super(delegate);
    }

    /**
     * Special constructor only intended for deserialization code.
     * 
     * @param charset
     *            encoding used for encoding/decoding strings.
     */
    private ChannelAccessClassNameImpl(Charset charset) {
        super(charset);
    }

    @Override
    public ChannelAccessValueType getType() {
        return ChannelAccessValueType.DBR_CLASS_NAME;
    }

    @Override
    public ChannelAccessClassName asReadOnlyValue() {
        if (delegate != null) {
            return this;
        }
        return new ChannelAccessClassNameImpl(this);
    }

    @Override
    public void serialize(ByteSink byteSink, int count) {
        if (delegate != null) {
            delegate.serialize(byteSink, count);
            return;
        }
        assert (count >= 0);
        assert (count <= value.size());
        serializeValue(byteSink, count);
    }

    @Override
    public boolean equals(Object obj) {
        if (delegate != null) {
            return delegate.equals(obj);
        }
        obj = extractDelegate(obj);
        return super.equals(obj);
    }

    @Override
    public int hashCode() {
        if (delegate != null) {
            return delegate.hashCode();
        }
        return super.hashCode();
    }

    @Override
    public ChannelAccessClassName clone() {
        if (delegate != null) {
            return delegate.clone();
        }
        ChannelAccessClassNameImpl copy = (ChannelAccessClassNameImpl) super
                .clone();
        return copy;
    }

    /**
     * Creates a DBR_CLASS_NAME from its serialized representation.
     * 
     * @param byteSource
     *            byte source to read data from.
     * @param count
     *            number of value elements to read.
     * @param charset
     *            character-set used for encoding and decoding strings.
     * @return DBR_CLASS_NAME instance created from the serialized data read
     *         from <code>byteSource</code>.
     * @throws BufferUnderflowException
     *             if the byte source throws such an exception.
     */
    public static ChannelAccessClassNameImpl deserialize(ByteSource byteSource,
            int count, Charset charset) {
        ChannelAccessClassNameImpl value = new ChannelAccessClassNameImpl(
                charset);
        value.deserializeValue(byteSource, count);
        return value;
    }

}
