/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import static org.junit.Assert.*;

import org.junit.Test;

import com.aquenos.epics.jackie.common.util.IntegerIdPool.Id;

/**
 * Tests for {@link IntegerIdPool}. These test only test the functionality
 * provided by the wrapper. Detailed tests including concurrency tests are made
 * by {@link BasicIntegerIdPoolTest}.
 * 
 * @author Sebastian Marsching
 */
public class IntegerIdPoolTest {

    /**
     * Test whether trying to get the numeric identifer after releasing an ID
     * causes an exception.
     */
    @Test(expected = IllegalStateException.class)
    public void getAfterRelease() {
        IntegerIdPool idPool = new IntegerIdPool(0, Integer.MAX_VALUE, 0L,
                false);
        Id id = idPool.acquireId();
        id.release();
        id.get();
    }

    /**
     * Test whether we can get the numeric identifier of an ID.
     */
    @Test
    public void getNumericId() {
        IntegerIdPool idPool = new IntegerIdPool(0, Integer.MAX_VALUE, 0L,
                false);
        Id id = idPool.acquireId();
        assertEquals(0, id.get());
    }

    /**
     * Test whether the numeric ID is reused after the ID has been released.
     */
    @Test
    public void reuseAfterRelease() {
        IntegerIdPool idPool = new IntegerIdPool(0, Integer.MAX_VALUE, 0L,
                false);
        Id id = idPool.acquireId();
        assertEquals(0, id.get());
        // Double release should not be an error, the ID should still only be
        // released once.
        id.release();
        id.release();
        assertFalse(id.isAlive());
        id = idPool.acquireId();
        // We expect the numeric ID to be reused.
        assertEquals(0, id.get());
        // If we do not release the ID, we should get another numeric ID.
        Id id2 = idPool.acquireId();
        assertEquals(1, id2.get());
    }

    /**
     * Tests the {@link Id}'s <code>toString()</code> method.
     */
    @Test
    public void idToString() {
        IntegerIdPool idPool = new IntegerIdPool(0, Integer.MAX_VALUE, 0L,
                false);
        Id id = idPool.acquireId();
        assertEquals("0", id.toString());
        id.release();
        assertEquals("<released>", id.toString());
    }

}
