/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import static org.junit.Assert.*;

import java.util.LinkedList;

import org.junit.Test;

/**
 * Tests for {@link LinkedWeakIdentityHashSet}.
 * 
 * @author Sebastian Marsching
 */
public class LinkedWeakIdentityHashSetTest {

    /**
     * This tests adds elements to the set without retaining a strong reference
     * to them. It checks that the number of elements is less than or equal to
     * the number of elements that have been added. After that, a a major GC is
     * triggered and it is checked that the size has not increased. We cannot
     * check whether the size has actually decreased, because there is no
     * guarantee that a GC run will remove elements that do not have strong
     * references.
     */
    @Test
    public void testManyElementsWithWeakReferenceOnly() {
        int numberOfElements = 2000000;
        LinkedWeakIdentityHashSet<Integer> set = new LinkedWeakIdentityHashSet<Integer>(
                numberOfElements);
        assertEquals(0, set.size());
        for (int i = 0; i < numberOfElements; i++) {
            set.add(new Integer(i));
        }
        int size1 = set.size();
        assertTrue(size1 <= numberOfElements);
        System.gc();
        int size2 = set.size();
        assertTrue(size2 <= size1);
    }

    /**
     * This test adds elements to the set and retains strong references to them.
     * It checks that the number of elements stays constant, even after running
     * a major GC.
     */
    @Test
    public void testManyElementsWithStrongReference() {
        int numberOfElements = 2000000;
        LinkedWeakIdentityHashSet<Integer> set = new LinkedWeakIdentityHashSet<Integer>(
                numberOfElements);
        // We add all values to a list to avoid them being garbage collected.
        LinkedList<Integer> list = new LinkedList<Integer>();
        assertEquals(0, set.size());
        for (int i = 0; i < numberOfElements; i++) {
            Integer value = new Integer(i);
            set.add(value);
            list.add(value);
        }
        assertEquals(numberOfElements, set.size());
        System.gc();
        assertEquals(numberOfElements, set.size());
        // Verify that the elements are returned in the correct order.
        int count = 0;
        for (int element : set) {
            assertEquals(count, element);
            count++;
        }
        // We have to use list again. Otherwise, it might get garbage collected
        // earlier and this would cause elements from set to be garbage
        // collected.
        assertEquals(numberOfElements, list.size());
    }

    /**
     * Test that elements are ordered correctly when removing and reinserting
     * them.
     */
    @Test
    public void orderTest() {
        Integer i1 = new Integer(1);
        Integer i2 = new Integer(2);
        Integer i3 = new Integer(3);
        Integer i4 = new Integer(4);
        LinkedWeakIdentityHashSet<Integer> set = new LinkedWeakIdentityHashSet<Integer>();
        set.add(i1);
        set.add(i2);
        set.add(i3);
        set.add(i4);
        LinkedList<Integer> testList1 = new LinkedList<Integer>();
        testList1.add(i1);
        testList1.add(i2);
        testList1.add(i3);
        testList1.add(i4);
        assertEquals(testList1, new LinkedList<Integer>(set));
        set.remove(i3);
        LinkedList<Integer> testList2 = new LinkedList<Integer>();
        testList2.add(i1);
        testList2.add(i2);
        testList2.add(i4);
        assertEquals(testList2, new LinkedList<Integer>(set));
        set.add(i3);
        LinkedList<Integer> testList3 = new LinkedList<Integer>();
        testList3.add(i1);
        testList3.add(i2);
        testList3.add(i4);
        testList3.add(i3);
        assertEquals(testList3, new LinkedList<Integer>(set));
        set.remove(i1);
        LinkedList<Integer> testList4 = new LinkedList<Integer>();
        testList4.add(i2);
        testList4.add(i4);
        testList4.add(i3);
        assertEquals(testList4, new LinkedList<Integer>(set));
        set.remove(i3);
        LinkedList<Integer> testList5 = new LinkedList<Integer>();
        testList5.add(i2);
        testList5.add(i4);
        assertEquals(testList5, new LinkedList<Integer>(set));
        set.add(i2);
        assertEquals(testList5, new LinkedList<Integer>(set));
        // Use testList1 again to make sure that the integers are not garbage
        // collected too early.
        assertEquals(4, testList1.size());
    }

}
