/*
 * Copyright 2015-2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client;

import java.util.Collection;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;

import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.util.ListenableFuture;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmAcknowledgementStatus;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessAlarmSeverity;
import com.aquenos.epics.jackie.common.value.ChannelAccessClassName;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessControlsShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessSimpleOnlyString;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeChar;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeDouble;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeFloat;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeLong;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeShort;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeString;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Representation of a Channel Access channel. A channel represents a process
 * variable that can be accessed via the Channel Access protocol.
 * </p>
 * 
 * <p>
 * Channels can be retrieved from a {@link ChannelAccessClient}. A channel
 * supports three different kinds of operations:
 * </p>
 * 
 * <ul>
 * <li>get (read),</li>
 * <li>put (write), and</li>
 * <li>monitor (subscription that receives updates on value changes).</li>
 * </ul>
 * 
 * <p>
 * Once created, a channel can be used until it is {@linkplain #destroy()
 * destroyed}. A channel keeps strong references to its
 * {@link ChannelAccessConnectionListener ChannelAccessConnectionListeners} and
 * to the {@link ChannelAccessClient}. However, depending on the implementation,
 * it might keep strong or weak references to its {@link ChannelAccessMonitor
 * ChannelAccessMonitors}. {@link ChannelAccessMonitor ChannelAccessMonitors}
 * keep a strong reference to their channel.
 * </p>
 * 
 * <p>
 * The {@link ChannelAccessClient} may keep strong or weak references to its
 * channels. Therefore, code using a channel (e.g. by registering a
 * {@link ChannelAccessConnectionListener} with it) should always keep a strong
 * reference to this channel in order to keep it alive. This strong reference
 * cannot be kept through the listener (unless an external strong reference to
 * the listener is kept) because this will not avoid garbage collection of the
 * channel. However, you still should call the channel's {@link #destroy()}
 * method if you do not need the channel any longer, because some
 * implementations might choose to keep a strong reference to the channel and in
 * this case the channel will not be destroyed unless you destroy it explicitly.
 * </p>
 * 
 * <p>
 * It is safe to create a separate channel instance for every place where one is
 * needed. Channel instances that refer to the same channel name (and belong to
 * the same {@link ChannelAccessClient}) will actually share one connection, so
 * that there is no significant overhead in having many channel objects for the
 * same channel. Having a separate channel instance for each purpose makes it
 * easier to decide when a channel is not needed any longer and can be
 * destroyed.
 * </p>
 * 
 * <p>
 * Please note that destroying a channel will also destroy all its monitors and
 * destroying a {@link ChannelAccessClient} will destroy all its channels (and
 * thus all their monitors).
 * </p>
 * 
 * <p>
 * Due to their asynchronous nature, the <code>get</code> and <code>put</code>
 * operations supplied by this interface do not return direct results. Instead,
 * {@link ListenableFuture}s that allow to wait for a result or to be notified
 * when a result is available are supplied.
 * </p>
 * 
 * <p>
 * These futures' {@link ListenableFuture#get() get} methods might throw an
 * {@link ExecutionException} that wraps a {@link ChannelAccessException}. This
 * happens if the error cannot be detected before the operation is started.
 * Otherwise, the method that starts the operation will throw the exception
 * directly.
 * </p>
 * 
 * <p>
 * Monitors, on the other hand, can always be created and signal error
 * conditions by calling the
 * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
 * monitorError(...)} method of their listeners.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ChannelAccessChannel {

    /**
     * <p>
     * Destroys this channel. The channel cannot be used any longer. This will
     * also destroy all monitors that belong to this channel. This method should
     * be called when the channel is not needed any longer.
     * </p>
     * 
     * <p>
     * Calling this method will not destroy other channel instances that have
     * been created by separate calls to
     * {@link ChannelAccessClient#getChannel(String)}, even if they refer to the
     * same channel.
     * </p>
     */
    void destroy();

    /**
     * Returns the name of this channel. The name identifies the channel in the
     * network.
     * 
     * @return channel name.
     */
    String getName();

    /**
     * Returns the Channel Access client this channel belongs to. This is the
     * client that was used to create this channel.
     * 
     * @return Channel Access client this channel belongs to.
     */
    ChannelAccessClient getClient();

    /**
     * <p>
     * Returns the native element count of this channel. The native element
     * count is the number of element that the server sends when the channel is
     * connected.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return native element count reported by the server.
     * @throws IllegalStateException
     *             if this channel is not connected ( {@link #isConnected()}
     *             returns <code>false</code>).
     */
    int getNativeCount();

    /**
     * <p>
     * Returns the native data type of this channel. The native data type is the
     * data type that the server sends when the channel is connected.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute change throughout the life-cycle of the channel. For a
     * disconnected channel, this method throws an {@link IllegalStateException}
     * .
     * </p>
     * 
     * @return native data type reported by the server.
     * @throws IllegalStateException
     *             if this channel is not connected ( {@link #isConnected()}
     *             returns <code>false</code>).
     */
    ChannelAccessValueType getNativeDataType();

    /**
     * Tells whether this channel is currently connected.
     * 
     * @return <code>true</code> if the channel is currently connected,
     *         <code>false</code> if it is currently disconnected.
     * @see #isConnectedOnce()
     */
    boolean isConnected();

    /**
     * Tells whether this channel has been connected at least once.
     * 
     * @return <code>true</code> if this channel is currently connected or has
     *         previously been connected, <code>false</code> otherwise.
     * @see #isConnected()
     */
    boolean isConnectedOnce();

    /**
     * <p>
     * Tells whether this channel has been destroyed. A channel is destroyed
     * when its {@link #destroy()} method is called or when its
     * {@linkplain #getClient() client} is destroyed. It might also be destroyed
     * if it is not used any longer (there are no strong references to it and it
     * is garbage collected).
     * </p>
     * 
     * <p>
     * After a channel has been destroyed, it cannot be used any longer and most
     * of its methods will throw an exception.
     * </p>
     * 
     * @return <code>true</code> if this channel has been destroyed,
     *         <code>false</code> if it is still alive.
     */
    boolean isDestroyed();

    /**
     * <p>
     * Tells whether this client is allowed to read from this channel. If the
     * client is allowed to read, the get and monitor operations can be used.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return <code>true</code> if this client is allowed to read from this
     *         channel and <code>false</code> if it not allowed to use read
     *         operations on this channel.
     * @throws IllegalStateException
     *             if this channel is not connected ({@link #isConnected()}
     *             returns <code>false</code>).
     */
    boolean isMayRead();

    /**
     * <p>
     * Tells whether this client is allowed to write this channel. If the client
     * is allowed to write, the put operations can be used.
     * </p>
     * 
     * <p>
     * As a channel might be disconnected and connect to a different server,
     * this attribute might change throughout the life-cycle of the channel. For
     * a disconnected channel, this method throws an
     * {@link IllegalStateException}.
     * </p>
     * 
     * @return <code>true</code> if this client is allowed to write to this
     *         channel and <code>false</code> if it not allowed to use write
     *         operations on this channel.
     * @throws IllegalStateException
     *             if this channel is not connected ({@link #isConnected()}
     *             returns <code>false</code>).
     */
    boolean isMayWrite();

    /**
     * <p>
     * Registers a connection state listener with this channel. The listener
     * will be notified when this channel's connection state changes. When the
     * listener is added, it is called once to notify it of the current
     * connection state.
     * </p>
     * 
     * <p>
     * If the specified listener is already registered with this channel, no
     * action is taken and this method returns <code>false</code>.
     * </p>
     * 
     * <p>
     * <strong>Important note</strong>: Typically, the listener is executed by
     * the thread that performs network I/O. For this reason, it is important
     * that the listener does not block or perform any long-running operations.
     * Instead, the listener should delegate any such operation to a different
     * thread (e.g. using an {@link Executor}). Any violation of this rule might
     * result in unexpected or unpredictable behavior. In particular, a blocking
     * listener might cause a dead-lock, which results in the whole Channel
     * Access client being stalled indefinitely.
     * </p>
     * 
     * <p>
     * If the calling code is not interested in receiving permanent status
     * updates and only wants to wait for a certain connection state to be
     * established, {@link #waitForConnectionState(boolean)} might be a simpler
     * alternative to this method.
     * </p>
     * 
     * @param listener
     *            listener to be registered with this channel. Must not be
     *            <code>null</code>.
     * @return <code>true</code> if the listener has been added,
     *         <code>false</code> if the listener has already been added
     *         earlier.
     * @see #removeConnectionListener(ChannelAccessConnectionListener)
     * @see #waitForConnectionState(boolean)
     * @throws NullPointerException
     *             if <code>listener</code> is null.
     */
    boolean addConnectionListener(ChannelAccessConnectionListener listener);

    /**
     * <p>
     * Unregisters a connection state listener from this channel. After being
     * removed, the listener will not be notified of further connection state
     * changes.
     * </p>
     * 
     * <p>
     * If the specified listener has already been removed from this channel or
     * has never been added, no action is taken and this method returns
     * <code>false</code>.
     * </p>
     * 
     * @param listener
     *            listener to be unregistered from this channel.
     * @return <code>true</code> if the listener has been removed,
     *         <code>false</code> if the listener has already been removed
     *         earlier or if it has never been added.
     * @see #addConnectionListener(ChannelAccessConnectionListener)
     */
    boolean removeConnectionListener(ChannelAccessConnectionListener listener);

    /**
     * <p>
     * Returns a listenable future that completes when the specified connection
     * state is established. This method can be used if certain actions are to
     * be delayed until the channel is connected or disconnected.
     * </p>
     * 
     * <p>
     * If the channel is already in the specified connection state, the future
     * returned by this method completes immediately.
     * </p>
     * 
     * <p>
     * Due to the asynchronous nature of connection events, there is no
     * guarantee that the channel is still in the specified connection state
     * when the future has completed. It is possible that the channel again
     * changed its state after first switching to the desired state. If the code
     * waiting for the future detects such a situation, it should simply make
     * another call to this method to wait for the connection state to change
     * again.
     * </p>
     * 
     * @param connectionState
     *            desired connection state. If <code>true</code>, the future
     *            returned by this method completes when the channel is
     *            connected. If <code>false</code>, the future returned by this
     *            method completes when the channel is disconnected.
     * @return listenable future that completes when the channel changes to the
     *         specified <code>connectionState</code>. If the channel already is
     *         in the specified <code>connectionState</code>, the future
     *         completes immediately. The completed future's value will be the
     *         connection state at the time when it completed (always the
     *         specified <code>connectionState</code>).
     */
    ListenableFuture<Boolean> waitForConnectionState(boolean connectionState);

    /**
     * <p>
     * Gets the current value from the channel using the specified type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be returned by the server. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes. The
     *         type of the value returned by the future can be expected to match
     *         the specified <code>valueType</code>.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType);

    /**
     * <p>
     * Gets the current value from the channel using the specified type and
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be returned by the server. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes. The
     *         type of the value returned by the future can be expected to match
     *         the specified <code>valueType</code>.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the specified <code>count</code> is
     *             negative.
     * @throws IllegalStateException
     *             if this channel has already been {@link #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<? extends ChannelAccessGettableValue<?>> get(
            ChannelAccessValueType valueType, int count);

    /**
     * <p>
     * Gets the current value from the channel using the channel's
     * {@linkplain #getNativeDataType() native data type} and
     * {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative();

    /**
     * <p>
     * Gets the current value from the channel using the channel's
     * {@linkplain #getNativeDataType() native data type} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<? extends ChannelAccessGettableValue<?>> getNative(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyString> getString();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyString> getString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getAlarmString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeString> getTimeString();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeString> getTimeString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getGraphicsString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyString> getControlsString(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyShort> getShort();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyShort> getShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyShort> getAlarmShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeShort> getTimeShort();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeShort> getTimeShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyShort> getGraphicsShort(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsShort> getControlsShort();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsShort> getControlsShort(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyFloat> getFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyFloat> getAlarmFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeFloat> getTimeFloat();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeFloat> getTimeFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyFloat> getGraphicsFloat(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsFloat> getControlsFloat();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsFloat> getControlsFloat(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyEnum> getEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyEnum> getAlarmEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeEnum> getTimeEnum();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeEnum> getTimeEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getGraphicsEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsEnum> getControlsEnum(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyChar> getChar();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyChar> getChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyChar> getAlarmChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeChar> getTimeChar();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeChar> getTimeChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyChar> getGraphicsChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsChar> getControlsChar();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsChar> getControlsChar(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyLong> getLong();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyLong> getLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyLong> getAlarmLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeLong> getTimeLong();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeLong> getTimeLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyLong> getGraphicsLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsLong> getControlsLong();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsLong> getControlsLong(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type and the
     * channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} and the specified
     * element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessSimpleOnlyDouble> getDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmOnlyDouble> getAlarmDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeDouble> getTimeDouble();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessTimeDouble> getTimeDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type and
     * the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessGraphicsOnlyDouble> getGraphicsDouble(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsDouble> getControlsDouble();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessControlsDouble> getControlsDouble(int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type and the channel's {@linkplain #getNativeCount() native element
     * count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} and
     * the specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessAlarmAcknowledgementStatus> getAlarmAcknowledgementStatus(
            int count);

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type
     * and the channel's {@linkplain #getNativeCount() native element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessClassName> getClassName();

    /**
     * <p>
     * Gets the current value from the channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} and the
     * specified element count}.
     * </p>
     * 
     * <p>
     * Due to a get operation's asynchronous nature, the result is not returned
     * right away, but a {@linkplain ListenableFuture future} that can be used
     * to wait on the operation and get its result is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the get operation only returns the number of elements
     *            that is equal to the channel's native element count.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the get operation cannot be started (e.g. because the
     *             channel is not connected or the operation is not allowed).
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<ChannelAccessClassName> getClassName(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The type of the values supplied with monitor
     *         events can be expected to match the specified
     *         <code>valueType</code>.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType);

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The type
     *         of the values supplied with monitor events can be expected to
     *         match the specified <code>valueType</code>.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The type of the values supplied with monitor
     *         events can be expected to match the specified
     *         <code>valueType</code> and the values can be expected to not
     *         exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation or if the specified <code>count</code> is
     *             negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count);

    /**
     * <p>
     * Creates a monitor for this channel using the specified type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param valueType
     *            type of the value that shall be sent for monitor events. The
     *            specified type must be allowed for get operations (
     *            {@link ChannelAccessValueType#isGettable()} must return
     *            <code>true</code>). If <code>null</code>, the channel's
     *            {@linkplain #getNativeDataType() native data type} is used.
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The type
     *         of the values supplied with monitor events can be expected to
     *         match the specified <code>valueType</code> and the values can be
     *         expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>valueType</code> is not allowed for a
     *             get operation, if the specified <code>count</code> is
     *             negative, or if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitor(
            ChannelAccessValueType valueType, int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the channel's
     * {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native data type and native element
     * count might vary over time, the type of the values and the number of
     * elements in the values of monitor events might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative();

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the channel's
     * {@linkplain #getNativeCount() native element count}, and the specified
     * event mask. As a channel's native data type and native element count
     * might vary over time, the type of the values and the number of elements
     * in the values of monitor events might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the specified element
     * count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native data type might vary over
     * time, the type of the values of monitor events might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the channel's
     * {@linkplain #getNativeDataType() native data type}, the specified element
     * count, and the specified event mask. As a channel's native data type
     * might vary over time, the type of the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<? extends ChannelAccessGettableValue<?>> monitorNative(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyString> monitorString(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_STRING DBR_STS_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorAlarmString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_STRING DBR_TIME_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeString> monitorTimeString(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_STRING DBR_GR_STRING} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorGraphicsString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_STRING DBR_CTRL_STRING} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyString> monitorControlsString(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyShort> monitorShort(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_SHORT DBR_STS_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyShort> monitorAlarmShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_SHORT DBR_TIME_SHORT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeShort> monitorTimeShort(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_SHORT DBR_GR_SHORT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyShort> monitorGraphicsShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_SHORT DBR_CTRL_SHORT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsShort> monitorControlsShort(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyFloat> monitorFloat(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_FLOAT DBR_STS_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyFloat> monitorAlarmFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_FLOAT DBR_TIME_FLOAT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeFloat> monitorTimeFloat(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_FLOAT DBR_GR_FLOAT} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyFloat> monitorGraphicsFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_FLOAT DBR_CTRL_FLOAT} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsFloat> monitorControlsFloat(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the specified
     * event mask. As a channel's native element count might vary over time, the
     * number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyEnum> monitorEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_ENUM DBR_STS_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyEnum> monitorAlarmEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_ENUM DBR_TIME_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeEnum> monitorTimeEnum(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_ENUM DBR_GR_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorGraphicsEnum(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_ENUM DBR_CTRL_ENUM} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsEnum> monitorControlsEnum(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the specified
     * event mask. As a channel's native element count might vary over time, the
     * number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyChar> monitorChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_CHAR DBR_STS_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyChar> monitorAlarmChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_CHAR DBR_TIME_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeChar> monitorTimeChar(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_CHAR DBR_GR_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyChar> monitorGraphicsChar(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_CHAR DBR_CTRL_CHAR} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsChar> monitorControlsChar(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the channel's
     * {@linkplain #getNativeCount() native element count}, and the specified
     * event mask. As a channel's native element count might vary over time, the
     * number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the specified
     * element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG} data type, the specified
     * element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyLong> monitorLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_LONG DBR_STS_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyLong> monitorAlarmLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_LONG DBR_TIME_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeLong> monitorTimeLong(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_LONG DBR_GR_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyLong> monitorGraphicsLong(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTLR_LONG} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_LONG DBR_CTRL_LONG} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsLong> monitorControlsLong(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessSimpleOnlyDouble> monitorDouble(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STS_DOUBLE DBR_STS_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmOnlyDouble> monitorAlarmDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_TIME_DOUBLE DBR_TIME_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessTimeDouble> monitorTimeDouble(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * channel's {@linkplain #getNativeCount() native element count}, and the
     * specified event mask. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_GR_DOUBLE DBR_GR_DOUBLE} data type, the
     * specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessGraphicsOnlyDouble> monitorGraphicsDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CTRL_DOUBLE DBR_CTRL_DOUBLE} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessControlsDouble> monitorControlsDouble(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the channel's {@linkplain #getNativeCount() native element count},
     * and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the channel's {@linkplain #getNativeCount() native element count},
     * and the specified event mask. As a channel's native element count might
     * vary over time, the number of elements in the values of monitor events
     * might vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_STSACK_STRING DBR_STSACK_STRING} data
     * type, the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessAlarmAcknowledgementStatus> monitorAlarmAcknowledgementStatus(
            int count, ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}. As a channel's native element count might vary over
     * time, the number of elements in the values of monitor events might vary
     * accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessClassName> monitorClassName();

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the channel's {@linkplain #getNativeCount() native element count}, and
     * the specified event mask. As a channel's native element count might vary
     * over time, the number of elements in the values of monitor events might
     * vary accordingly.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask.
     * @throws IllegalArgumentException
     *             if the <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the specified element count, and the
     * {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     * default event mask}.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @return monitor for this channel using the
     *         {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *         default event mask}. The values supplied with monitor events can
     *         be expected to not exceed the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(int count);

    /**
     * <p>
     * Creates a monitor for this channel using the
     * {@link ChannelAccessValueType#DBR_CLASS_NAME DBR_CLASS_NAME} data type,
     * the specified element count, and the specified event mask.
     * </p>
     * 
     * <p>
     * A monitor can be created for a channel that is currently disconnected (or
     * has not yet been connected). However, the monitor will only post events
     * when the channel is connected.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by calling the monitor listeners'
     * {@link ChannelAccessMonitorListener#monitorError(ChannelAccessMonitor, com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus, String)
     * monitorError(...)} method. Only errors in the parameters to this method
     * or a destroyed channel will result in an immediate exception thrown by
     * this method.
     * </p>
     * 
     * @param count
     *            number of elements to request from the server. Must be
     *            positive or zero. A value of zero is interpreted as the
     *            channel's {@linkplain #getNativeCount() native element count}.
     *            If the specified number exceeds the channel's native element
     *            count, the monitor event will only provide the number of
     *            elements that is equal to the channel's native element count.
     * @param eventMask
     *            event mask to use for the monitor. The event mask specifies
     *            which events on the server will trigger a monitor event. If
     *            <code>null</code>, the
     *            {@linkplain ChannelAccessClientConfiguration#getDefaultEventMask()
     *            default event mask} is used.
     * @return monitor for this channel using the specified event mask. The
     *         values supplied with monitor events can be expected to not exceed
     *         the specified number of elements.
     * @throws IllegalArgumentException
     *             if the specified <code>count</code> is negative or if the
     *             <code>eventMask</code> is
     *             {@link ChannelAccessEventMask#DBE_NONE}.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ChannelAccessMonitor<ChannelAccessClassName> monitorClassName(int count,
            ChannelAccessEventMask eventMask);

    /**
     * <p>
     * Puts the specified value to the channel. The value can be any puttable
     * value. However, specifying an element count that is larger than the
     * channel's {@linkplain #getNativeCount() native element count} might
     * result in an error. However, most servers will simply ignore the extra
     * element's and only use the beginning of the array.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> put(ChannelAccessPuttableValue<?> value);

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putString(String value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is copied, so subsequent changes to the
     * collection will not have any effect on the put operation.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     */
    ListenableFuture<Void> putString(String[] value);

    /**
     * <p>
     * Puts the specified collection to the channel as a
     * {@link ChannelAccessValueType#DBR_STRING DBR_STRING}. Specifying a
     * collection that is larger than the channel's
     * {@linkplain #getNativeCount() native element count} might result in an
     * error. However, most servers will simply ignore the extra element's and
     * only use the beginning of the array. The specified collection is copied,
     * so subsequent changes to the collection will not have any effect on the
     * put operation.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code> or contains
     *             <code>null</code> elements.
     */
    ListenableFuture<Void> putString(Collection<String> value);

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_SHORT
     * DBR_SHORT} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putShort(short value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_SHORT DBR_SHORT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putShort(short[] value);

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_FLOAT
     * DBR_FLOAT} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putFloat(float value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_FLOAT DBR_FLOAT}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putFloat(float[] value);

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_ENUM
     * DBR_ENUM} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putEnum(short value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_ENUM DBR_ENUM}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putEnum(short[] value);

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_CHAR
     * DBR_CHAR} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putChar(byte value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_CHAR DBR_CHAR}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putChar(byte[] value);

    /**
     * <p>
     * Puts the specified single element {@link ChannelAccessValueType#DBR_LONG
     * DBR_LONG} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putLong(int value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_LONG DBR_LONG}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putLong(int[] value);

    /**
     * <p>
     * Puts the specified single element
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE} to the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putDouble(double value);

    /**
     * <p>
     * Puts the specified array to the channel as a
     * {@link ChannelAccessValueType#DBR_DOUBLE DBR_DOUBLE}. Specifying an array
     * that is larger than the channel's {@linkplain #getNativeCount() native
     * element count} might result in an error. However, most servers will
     * simply ignore the extra element's and only use the beginning of the
     * array. The specified arrays is not copied, so the calling code should not
     * make any changes to the array after passing it to this method.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only errors in
     * the parameters to this method or a destroyed channel will result in an
     * immediate exception thrown by this method.
     * </p>
     * 
     * @param value
     *            value written to the channel.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalArgumentException
     *             if the <code>value</code> has zero elements.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if the <code>value</code> is <code>null</code>.
     */
    ListenableFuture<Void> putDouble(double[] value);

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKT
     * DBR_PUT_ACKT} to the channel. This can be used to configure the transient
     * alarm setting of the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param explicitlyAcknowledgeTransientAlarms
     *            <code>true</code> if transient alarms should have to be
     *            acknowledged explicitly, <code>false</code> if they should be
     *            automatically acknowledged when they disappear.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     */
    ListenableFuture<Void> putConfigureAcknowledgeTransientAlarms(
            boolean explicitlyAcknowledgeTransientAlarms);

    /**
     * <p>
     * Puts the specified {@link ChannelAccessValueType#DBR_PUT_ACKS
     * DBR_PUT_ACKS} to the channel. This can be used to acknowledge alarms for
     * the channel.
     * </p>
     * 
     * <p>
     * Due to a put operation's asynchronous nature, this method does not block,
     * but a {@linkplain ListenableFuture future} that can be used to wait on
     * the operation and to get an error (if there is any) is returned.
     * </p>
     * 
     * <p>
     * Errors are typically indicated by an exception that is wrapped in an
     * {@link ExecutionException} that is thrown by the future. Only a destroyed
     * channel will result in an immediate exception thrown by this method.
     * </p>
     * 
     * @param acknowledgeAlarmSeverity
     *            highest alarm severity that should be acknowledged (never
     *            <code>null</code>). This means that the alarm will be
     *            acknowledged if the severity of the unacknowledged alarm is
     *            less than or equal to the specified severity.
     * @return listenable future that can be used to retrieve the result of the
     *         operation and to be notified when the operation finishes.
     * @throws ChannelAccessException
     *             if the put operation cannot be started because the channel is
     *             not connected or there is a problem with the specified
     *             parameters.
     * @throws IllegalStateException
     *             if this channel has already been {@linkplain #isDestroyed()
     *             destroyed}.
     * @throws NullPointerException
     *             if <code>acknowledgeAlarmSeverity</code> is <code>null</code>
     *             .
     */
    ListenableFuture<Void> putAcknowledgeAlarm(
            ChannelAccessAlarmSeverity acknowledgeAlarmSeverity);

}
